import pandas as pd


def read_csv(*args, date_cols=None, **kwargs):
    """Read CSV file and convert date and numeric columns to the correct
    data types.

    Args:
        date_cols (list or None): if None, use *date_cols* ["obs_date", "creation_date",
            "modified_date", "collected_date"]; if "auto", use all columns ending with
            "_date"; otherwise a list of columns

    Other arguments and keyword arguments are passed to :func:`pandas.read_csv`.

    Returns:
        pandas DataFrame

    """
    df = pd.read_csv(*args, **kwargs)
    if date_cols == "auto":
        date_cols = [c for c in df.columns if c.endswith("_date")]
    elif date_cols is None:
        date_cols = ("obs_date", "creation_date", "modified_date", "collected_date")
    for col in date_cols:
        if col in df:
            df[col] = pd.to_datetime(df[col], errors="coerce")
    for col in df.columns:
        if not col in date_cols:
            df[col] = pd.to_numeric(df[col], errors="ignore")
    return df



import sys
from datetime import datetime
import logging
import os

import click
import matplotlib.pyplot as plt
import pandas as pd

import dew_gwdata as gd  # Access to SA Geodata database
import wrap_technote as tn  # GW data analysis functions
import ausweather

from pandas.plotting import register_matplotlib_converters

register_matplotlib_converters()


logger = tn.get_logger()

from ..charts_utils import apply_technote_chart_style

apply_technote_chart_style()


@click.command()
@click.option("-p", "--reporting-period", default=tn.CURRENT_RPERIOD)
@click.option("-v", "--verbose", count=True)
@click.option("--overwrite/--no-overwrite", default=True, help="True by default")
@click.option("--download/--no-download", default=False, help="False by default")
@click.option("--figures/--no-figures", default=True, help="True by default")
@click.option("--static/--no-static", default=True, help="True by default")
@click.option("--id-glob", default="*", help="match only one or two station IDs")
@click.option("--name-glob", default="*", help="match only one or two station names")
@click.option(
    "--internal-only/--no-internal-only", default=False, help="False by default"
)
@click.argument("report_key")
def rainfall(
    report_key,
    reporting_period,
    verbose,
    static,
    overwrite,
    download,
    figures,
    id_glob,
    name_glob,
    internal_only,
):
    return rainfall_run(
        report_key=report_key,
        reporting_period=reporting_period,
        verbose=verbose,
        static=static,
        overwrite=overwrite,
        download=download,
        figures=figures,
        id_glob=id_glob,
        name_glob=name_glob,
        internal_only=internal_only,
    )


def rainfall_run(
    report_key,
    reporting_period,
    verbose=0,
    static=True,
    overwrite=True,
    download=False,
    figures=True,
    id_glob="*",
    name_glob="*",
    internal_only=False,
):
    handlers = []
    if verbose == 1:
        handlers.append({"sink": sys.stdout, "level": "INFO"})
    if verbose == 2:
        handlers.append({"sink": sys.stdout, "level": "INFO"})
        handlers.append({"sink": "wraptn_rainfall.log", "level": "DEBUG"})
    config = {
        "handlers": handlers,
    }
    logger.configure(**config)

    logger.info(f"Using dew_gwdata {gd.__version__}")
    logger.info(f"Using wrap_technote {tn.__version__}")
    logger.info(f"Using ausweather {ausweather.__version__}")

    plt.rcParams["figure.dpi"] = 120

    logger.info(f"Loading report {report_key} for reporting_period {reporting_period}")
    logger.info(f"download={download} overwrite={overwrite} figures={figures}")

    report = tn.load_report(report_key, reporting_period)

    report.rainfall_path.mkdir(exist_ok=True, parents=True)

    # Download and store data.
    # logger.info(f"{report.path}")

    if download:
        logger.info("Downloading")
        for station_id in report.read_table("Report_Rainfall_stations").station_id:
            rrow = report.rainfall_dfn(station_id)
            if rrow.station_type == "BoM":
                logger.info(f"Downloading {station_id} from {rrow.data_start}")
                rf = tn.RainfallStationData.from_bom_via_silo(
                    station_id, data_start=rrow.data_start
                )

            elif rrow.station_type in ("Hydstra", "Aquarius"):
                rf = tn.RainfallStationData.from_aquarius(
                    station_id, data_start=rrow.data_start
                )

            rf.save_to_excel(report, overwrite=overwrite)
            rf.save_to_database(report)

            # data = ausweather.fetch_bom_station_from_silo(
            #     station_id, "groundwater@sa.gov.au", rrow.data_start
            # )
            # stub = f'{data["station_no"]}_{data["station_name"]}'
            # data["df"].loc[:, "wu_year"] = [
            #     tn.date_to_wateruseyear(d) for d in data["df"].Date
            # ]
            # data["wateruse_year"] = (
            #     data["df"].groupby("wu_year").Rain.sum().reset_index()
            # )

            # excel_filename = (
            #     report.rainfall_path / f"silo_data_{report.report_key}_{stub}.xlsx"
            # )
            # if overwrite:
            #     with pd.ExcelWriter(excel_filename) as writer:
            #         data["df"].to_excel(writer, "daily", index=False)
            #         data["annual"].reset_index().to_excel(
            #             writer, "annual (calendar)", index=False
            #         )
            #         data["srn"].reset_index().to_excel(
            #             writer, "annual provenance", index=False
            #         )
            #         data["wateruse_year"].reset_index().to_excel(
            #             writer, "annual (water-use year)", index=False
            #         )

    logger.info(f"Finding stations matching id_glob={id_glob} name_glob={name_glob}")
    stations = report.find_rainfall_stations(id_glob=id_glob, name_glob=name_glob)
    logger.info(f"Found {len(stations)} stations")
    dfs = {}
    dfns = {}
    for station in stations:
        logger.info(f"Loading data for {station['id']} {station['name']}")
        dfs[station["id"]] = report.get_all_rainfall_data(station)
        dfns[station["id"]] = report.rainfall_dfn(station["id"])

    # Make charts and spreadsheet.
    if figures:
        for station in stations:
            sid = station["id"]
            name = station["name"]

            logger.info(f"Making graphs for {sid} {name}")
            dfn = dfns[station["id"]]
            ann_calendar_df = dfs[sid]["annual (calendar)"]
            ann_srn_df = dfs[sid]["annual provenance"]
            ann_financial_df = dfs[sid]["annual (water-use year)"]
            daily_df = dfs[sid]["daily"]
            mp_df = daily_df[
                (daily_df["Date"] >= dfn.mean_period_start)
                & (daily_df["Date"] <= dfn.mean_period_finish)
            ]
            means = mp_df.groupby("wu_year").Rain.sum().describe()
            p_df = daily_df[
                (daily_df["Date"] >= dfn.rf_period_start)
                & (daily_df["Date"] <= dfn.rf_period_finish)
            ]
            rf_period_years = p_df.Date.dt.year.unique()
            if len(rf_period_years) == 1:
                year_type = "calendar"
            else:
                year_type = "split"
            logger.info(f"Selecting year_type = '{year_type}'")

            # for rf in report.rainfall_files():
            #     sid = rf["sid"]
            #     name = rf["name"]
            #     logger.info(f"{sid} {name}")
            #     rdfn = report.rainfall_dfn(sid)

            # Excel does not understand datetimes <= 1900. You will need to include the commented-out line
            # in that case.
            #     df['Date'] = pd.to_datetime(df['Date'], format='%Y-%m-%d %H:%M:%S')

            if not internal_only:
                # ----------------- SILO chart -------------------------
                logger.info(
                    f"Making chart :: ausweather.plot_silo_station :: for {sid} {name}"
                )
                pchart = ausweather.plot_silo_station(
                    ann_calendar_df.set_index("Date").Rain,
                    ann_calendar_df.Rain.mean(),
                    ann_srn_df.set_index("Date"),
                    title=f"{sid} {name}",
                )
                pchart_fn = report.rainfall_path / f"plot_station_{sid}_{name}.png"
                pchart["fig"].tight_layout()
                pchart["fig"].savefig(pchart_fn, dpi=150)

                # ---------------- Annual totals chart ----------------------
                # Mean period data

                # Annual water-use year chart
                # if year_type == "split":
                logger.info(
                    f"Making chart :: wrap_technote.plot_rainfall_wuy_annual :: for {sid} {name}"
                )
                filename = f"annual_rainfall_wuy_{sid}_{name}_{mp_df.iloc[0].wu_year}_{mp_df.iloc[-1].wu_year}.png"
                wuy_ann_chart = tn.plot_rainfall_wuy_annual(
                    mp_df, trend_line=True, wuy_col=False
                )
                wuy_ann_chart["ax"].figure.tight_layout()
                wuy_ann_chart["ax"].figure.savefig(
                    str(report.rainfall_path / filename), dpi=150
                )

                wuy_trend_line_df = pd.DataFrame(
                    [
                        {
                            "start": dfn.mean_period_start,
                            "finish": dfn.mean_period_finish,
                            "slope (mm/y)": wuy_ann_chart["slope"],
                            "intercept": wuy_ann_chart["intercept"],
                        }
                    ]
                )

                # if year_type == "calendar":
                # Annual calendar year chart
                logger.info(
                    f"Making chart :: wrap_technote.plot_rainfall_calendar_annual :: for {sid} {name}"
                )
                filename = f"annual_rainfall_calendar_{sid}_{name}_{mp_df.iloc[0].Date.year}_{mp_df.iloc[-1].Date.year}.png"
                calendar_ann_chart = tn.plot_rainfall_calendar_annual(
                    mp_df, trend_line=True, year_col=False
                )
                calendar_ann_chart["ax"].figure.tight_layout()
                calendar_ann_chart["ax"].figure.savefig(
                    str(report.rainfall_path / filename), dpi=150
                )

                calendar_trend_line_df = pd.DataFrame(
                    [
                        {
                            "start": dfn.mean_period_start,
                            "finish": dfn.mean_period_finish,
                            "slope (mm/y)": calendar_ann_chart["slope"],
                            "intercept": calendar_ann_chart["intercept"],
                        }
                    ]
                )

                # -------------------- Monthly totals (and means) chart -----------------------

                # Monthly means (mean_monthly) and monthly totals for last 18 months (period_monthly).
                if year_type == "split":
                    mlabel = (
                        tn.date_to_wateruseyear(dfn.mean_period_start)
                        + " to "
                        + tn.date_to_wateruseyear(dfn.mean_period_finish)
                    )
                    mp_label = f"{mp_df.iloc[0].wu_year}_{mp_df.iloc[-1].wu_year}"
                    plabel = f"{dfn.rf_period_start.year}-{str(dfn.rf_period_finish.year)[-2:]}"
                elif year_type == "calendar":
                    mlabel = (
                        f"{dfn.mean_period_start.year} to {dfn.mean_period_finish.year}"
                    )
                    mp_label = mlabel.replace(" to ", "_")
                    plabel = str(dfn.rf_period_start.year)

                xl_left = str(dfn.rf_period_start.year)
                xl_right = str(dfn.rf_period_finish.year)

                mean_monthly = mp_df.groupby(
                    [mp_df.Date.dt.year, mp_df.Date.dt.month]
                ).Rain.sum()
                mean_monthly = (
                    mean_monthly.reset_index(level=0, drop=True)
                    .reset_index()
                    .groupby("Date")
                    .mean()
                    .Rain
                )
                # mean_monthly = (
                #     mp_df.groupby([mp_df.Date.dt.year, mp_df.Date.dt.month])
                #     .Rain.sum()
                #     .mean(level=1)
                # )
                period_monthly = p_df.groupby(
                    [daily_df.Date.dt.year, daily_df.Date.dt.month]
                ).Rain.sum()

                # Monthly chart
                logger.info(
                    f"Making chart :: wrap_technote.plot_rainfall_monthly_means :: for {sid} {name} year_type={year_type}"
                )
                ax = tn.plot_rainfall_monthly_means(
                    mean_monthly,
                    period_monthly,
                    mlabel,
                    plabel,
                    xl_left,
                    xl_right,
                    year_type=year_type,
                )
                monthly_filename = (
                    f"monthly_rainfall_{sid}_{name}_"
                    + mp_label
                    + f"_period_{plabel}.png"
                )
                ax.figure.tight_layout()
                ax.figure.savefig(str(report.rainfall_path / monthly_filename), dpi=150)

            plt.style.use("default")
            # --------- Internal analysis -------
            internal_path = report.rainfall_path / "analysis_figures"
            internal_path.mkdir(parents=True, exist_ok=True)

            ax = tn.plot_annual_rainfall_internal(
                station,
                report,
                df=ann_calendar_df,
                chart_start=1990,
                sheet_name="calendar",
            )
            fn = f"annual_rainfall_internal_calendar_{sid}_{name}.png"
            ax.figure.tight_layout()
            ax.figure.savefig(report.rainfall_path / "analysis_figures" / fn, dpi=100)

            ax = tn.plot_annual_rainfall_internal(
                station,
                report,
                df=ann_financial_df,
                chart_start=1990,
                sheet_name="financial",
            )
            fn = f"annual_rainfall_internal_financial_{sid}_{name}.png"
            ax.figure.tight_layout()
            ax.figure.savefig(report.rainfall_path / "analysis_figures" / fn, dpi=100)

            current_year = dfn.rf_period_finish.year
            for chart_year in [current_year - i for i in range(4)]:
                chart_year = str(chart_year)
                ax = tn.plot_monthly_rainfall_internal(
                    station,
                    report,
                    show_year=chart_year,
                    dailydf=daily_df,
                )
                fn = f"monthly_rainfall_internal_{chart_year}_{sid}_{name}.png"
                ax.figure.tight_layout()
                ax.figure.savefig(
                    report.rainfall_path / "analysis_figures" / fn, dpi=100
                )

            ax = tn.plot_seasonal_rainfall_residuals_internal(
                station, report, daily_df=daily_df, chart_start=1990, value_type="mm"
            )
            fn = f"seasonal_rainfall_residuals_internal_mm_{sid}_{name}.png"
            ax.figure.tight_layout()
            ax.figure.savefig(report.rainfall_path / "analysis_figures" / fn, dpi=100)

            ax = tn.plot_seasonal_rainfall_residuals_internal(
                station, report, daily_df=daily_df, chart_start=1990, value_type="%"
            )
            fn = f"seasonal_rainfall_residuals_internal_pct_{sid}_{name}.png"
            ax.figure.tight_layout()
            ax.figure.savefig(report.rainfall_path / "analysis_figures" / fn, dpi=100)

            ax = tn.plot_seasonal_rainfall_totals_internal(
                station, report, daily_df=daily_df, chart_start=1990
            )
            fn = f"seasonal_rainfall_totals_internal_{sid}_{name}.png"
            ax.figure.tight_layout()
            ax.figure.savefig(report.rainfall_path / "analysis_figures" / fn, dpi=100)

            # ------------------------ Write data to spreadsheet -----------------------------

            if not internal_only:
                with pd.ExcelWriter(
                    str(
                        report.rainfall_path
                        / f"report_rainfall_{report.report_key}_{sid}_{name}_{mp_label}.xlsx"
                    )
                ) as writer:
                    wuy_ann_chart["rf_wuy"].to_excel(
                        writer, "water-use year", index=True
                    )
                    calendar_ann_chart["rf_year"].to_excel(
                        writer, "calendar", index=True
                    )
                    means.to_excel(writer, "means", index=True)
                    wuy_trend_line_df.to_excel(
                        writer, "trend line (water-use years)", index=True
                    )
                    calendar_trend_line_df.to_excel(
                        writer, "trend line (calendar years)", index=True
                    )
                    if year_type == "split":
                        wuy_trend_line_df.to_excel(writer, "trend line", index=True)
                    elif year_type == "calendar":
                        calendar_trend_line_df.to_excel(
                            writer, "trend line", index=True
                        )
                    if len(mean_monthly):
                        mean_monthly.to_excel(writer, f"monthly means", index=True)
                    if len(period_monthly):
                        period_monthly.to_excel(
                            writer, f"{plabel} monthly totals", index=True
                        )

    if static:
        create_static_rainfall_page(report)


def create_static_rainfall_page(report):
    templates = tn.load_html_templates()
    template = templates.get_template("static/rainfall.html")

    with open(str(report.rainfall_path / "rainfall.html"), mode="w") as html:
        data = report.calculated_rainfall_files()
        station_data = {k: v for k, v in data.items() if not k == "pivoted"}
        html.write(
            template.render(
                {"report": report, "data": station_data, "wide_data": data["pivoted"]}
            )
        )



from pathlib import Path
import logging

# import datatest
import dew_gwdata as gd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import ticker as mticker

# import pytest
from matplotlib import gridspec
import wrap_technote as tn

from scipy import optimize, stats

from .utils import *
from .gwutils import *
from .charts_utils import *
from .charts_wl import *

logger = get_logger()


def filter_wl_observations(df, return_removals=False, **kwargs):
    """Filter out water level observations which are not suitable for monitoring
    trend analysis.

    Args:
        df (pandas.DataFrame): water level observations
        return_removals (bool): if True, also return the records which were
            removed (and why)

    Returns:
        either a :class:`pandas.DataFrame` with the filtered water levels,
        or a tuple of two dataframes, where the second is the records which were
        removed.

    See :func:`load_qc_removals` for more keyword arguments.

    Which records does it drop?

    1. Drops rows where "Swl" is null
    2. Drops rows where "anomalous_ind" is not "N"
    3. Drops rows where "pumping_ind" is not "N"
    4. Drops rows where "measured_during" is either "D", "R", or "U"
    5. Drops rows where "dry_ind" is not "N"
    6. Drops rows where "comments" contains "[Missed peak recovery]"


    """
    if not "database" in df:
        df["database"] = "SA Geodata"

    df = df[~df.obs_date.isnull()]

    removals = [
        {
            "reason": "SA Geodata records with null SWL measurements",
            "idx": (df.database == "SA Geodata") & pd.isnull(df.swl),
        },
        {
            "reason": "SA Geodata records with '[Missed peak recovery]' in the comments field",
            "idx": df.comments.str.contains(
                "[Missed peak recovery]", regex=False, na=False
            ),
        },
        {
            "reason": "SA Geodata records with anomalous_ind != 'N'",
            "idx": (df.database == "SA Geodata") & (df.anomalous_ind != "N"),
        },
        {
            "reason": "SA Geodata records with pumping_ind != 'N'",
            "idx": (df.database == "SA Geodata") & (df.pumping_ind != "N"),
        },
        {
            "reason": "SA Geodata records with measured_during in 'D', 'R', 'U'",
            "idx": (df.database == "SA Geodata")
            & (df.measured_during.isin(["D", "R", "U"])),
        },
        {
            "reason": "SA Geodata records with dry_ind != 'N'",
            "idx": (df.database == "SA Geodata") & (df.dry_ind != "N"),
        },
        {
            "reason": "AQTS grade <= 0",
            "idx": (df.database == "Aquarius") & (df.grade <= 0),
        },
    ]

    removals += load_qc_removals(df, parameter="WL", **kwargs)

    rdfs = []
    comb_idx = np.array([False for x in range(len(df))])
    for removal in removals:
        idx = removal["idx"].values
        rdf = pd.DataFrame(df[idx])
        rdf["reason"] = removal["reason"]
        rdfs.append(rdf)
        comb_idx = np.logical_or(comb_idx, idx)

    df = df[~comb_idx]

    if return_removals:
        return df, pd.concat(rdfs)
    else:
        return df


def analyse_wl_by_seasons(df, seasons, well_id_col="dh_no", dt_col="obs_date"):
    """Reduce a water level dataset to one or more values per seasonal period,
    according to the definition of seasons in *seasons*.

    Args:
        df (:class:`pandas.DataFrame`): data table of water level observations
        well_id_col (``str``, optional): column of *df* which uniquely identifies
            an individual well - not needed if *df* only has data from one well.
        dt_col (``str``, optional): column of *df* with datetimes

    Returns:
        :class:`pandas.DataFrame`: dataframe with the same columns as *df* and the additional
        columns "season" and "year+season". The returned DataFrame can be used as
        the *anndf* argument to, for example, :func:`wrap_technote.plot_wl_seasonality`.

    Normally this would reduce df such that it has no more than one value per season.
    This is the case where the seasonal definition uses a reduction function such
    as 'max' or 'min'. However, when the reduction function is 'all', it
    will potentially retain more than one value per season. This is to accommodate
    regions such as Far North.

    """
    df2 = df.copy()
    df2.loc[:, "season"] = df2[dt_col].dt.dayofyear.map(seasons.dayofyear_to_season)
    df2.loc[:, "season_year"] = df2[dt_col].apply(seasons.label_year)
    df2.loc[:, "year+season"] = df2[dt_col].apply(seasons.label_year_and_season)
    df2.loc[:, "seasons_dfn"] = seasons.to_str()
    logger.debug(
        f"well_id: {df2[well_id_col].unique()[0]} seasons_dfn: id={id(seasons)} {seasons.to_str()}"
    )
    if well_id_col in df2.columns:
        groupby_cols = [well_id_col, "year+season", "season"]
    else:
        groupby_cols = ["year+season", "season"]
    if seasons.reduces_to_one_value_per_season:
        return (
            df2.groupby(groupby_cols)
            .apply(seasons.apply_season_func)
            .reset_index(drop=True)
        )
    else:
        return df2


def apply_calculate_seasonal_water_levels(resource):
    """Returns a function suitable for "apply"ing to a DataFrame.

    Args:
        resource (tn.Resource)

    Returns:
        function: a function that accepts a dataframe of water levels
        for a single well.

    The function can then be applied to a dataframe to produce a
    new dataframe containing seasonal water levels (using the function
    :func:`wrap_technote.analyse_wl_by_seasons`).

    """

    def apply_calculate_seasonal_water_levels__generated(frame):
        well_id = frame.well_id.unique()[0]
        season = resource.get_season(well_id)
        return analyse_wl_by_seasons(frame, season)

    return apply_calculate_seasonal_water_levels__generated


def get_total_wl_changes(wldata, windows):
    """
    df = resource.read_data("recovery_wl_data", "annual recovered WL")

    .. todo:: Document this function.

    """

    window_year_col = "season_year"

    dfs = []
    for window in windows:
        name = window[0]
        start_syears = list(window[1])
        start_syears += [f"{y}-{str(y + 1)[-2:]}" for y in start_syears]
        start_syears += [str(w) for w in window[1]]
        end_syears = list(window[2])
        end_syears += [f"{y}-{str(y + 1)[-2:]}" for y in end_syears]
        end_syears += [str(w) for w in window[2]]
        agg_arg = window[3]

        def get_start_end_year_rswls(f):
            start_rswl = f[f[window_year_col].isin(start_syears)].rswl.agg(agg_arg)
            end_rswl = f[f[window_year_col].isin(end_syears)].rswl.agg(agg_arg)
            return pd.Series({"start_window": start_rswl, "end_window": end_rswl})

        x = wldata.groupby("well_id").apply(get_start_end_year_rswls)
        x["change"] = x["end_window"] - x["start_window"]
        x.loc[x.change > 0, "change_direction"] = "increase"
        x.loc[x.change == 0, "change_direction"] = "no change"
        x.loc[x.change < 0, "change_direction"] = "decrease"
        x_cols = list(x.columns)
        x["window"] = name
        x["start_dfn"] = str(start_syears)
        x["end_dfn"] = str(end_syears)
        x = x[["window"] + x_cols + ["start_dfn", "end_dfn"]]
        dfs.append(x)

    return pd.concat(dfs).dropna(subset=["change"], how="any")


def get_total_wl_change_windows(current_year, offset_years=(30, 20, 10)):
    """

    .. todo:: Document this function.

    """
    windows = []
    for offset in offset_years:
        label = f"{offset} years"
        start_years = tuple(
            [
                current_year - offset,
                current_year - offset + 1,
                current_year - offset + 2,
            ]
        )
        end_year = tuple([current_year])
        func = "mean"
        windows.append((label, start_years, end_year, func))
    return windows


def collate_waterlevel_summary_data(resource, db=None):
    """Collate water level data outputs for a resource.

    Args:
        resource (Resource object)
        db (SAGeodataConnection object): optional, will
            be created if omitted.

    Returns:
        dict: details below

    The keys of the returned dictionary are:

    - unlisted_wells
    - curr_rank_wells
    - curr_rank_excluded_wells
    - final_trend_wells
    - excluded_trend_wells
    - resource: wrap_technote.Resource object
    - wells: list of dicts
    - wells_html: HTML representation
    - data_val
    - curr_ranks
    - curr_ranks_summ
    - curr_ranks_summ_grpd
    - majority_grpd_curr_ranks
    - curr_ranks_qc
    - final_trends
    - wltrends_summ_by_triclass
    - wltrends_summ_all
    - total_wl_changes
    - total_wl_changes_summ_1
    - total_wl_changes_summ_2

    """
    if db is None:
        db = gd.sageodata()

    s1 = resource.read_data("validated_data", sheet_name=None)
    s2 = resource.read_data("recovery_wl_data", sheet_name=None)

    ranks = s2["annual_recovered WL"]
    curr_ranks = s2["current ranked WLs"]
    curr_ranks_excluded = s2["current ranks excl"]
    curr_ranks_qc = s2["data quality"]

    s3 = resource.read_data("recovery_wl_trends", sheet_name=None)
    final_trends = s3["final trends"]
    excluded_trends = s3["excluded trends"]

    total_wl_changes = s2["total_wl_changes"]
    curr_rank_wells = sorted(set(curr_ranks.well_id.unique()))
    curr_rank_excluded_wells = sorted(set(curr_ranks_excluded.well_id.unique()))
    final_trend_wells = sorted(set(final_trends.well_id.unique()))
    excluded_trend_wells = sorted(set(excluded_trends.well_id.unique()))

    all_listed_wells = set(
        sorted(
            set(
                list(curr_rank_wells)
                + list(final_trend_wells)
                + list(curr_rank_excluded_wells)
                + list(excluded_trend_wells)
            )
        )
    )
    all_wells = set(
        list(s1["valid data"].well_id.unique())
        + list(s1["invalid data"].well_id.unique())
    )
    unlisted_wells = all_wells - set(all_listed_wells)

    wells = db.drillhole_details(db.find_wells(str(all_wells))).sort_values("well_id")
    wells["charts_link"] = wells.apply(
        lambda x: f"<a href='#charts-{x.well_id}'>{x.well_id}</a>", axis="columns"
    )

    data_val = resource.filter_table_by_well_ids(
        wells.well_id.values, table="Data_validation"
    )
    data_val_cols = [
        "well_id",
        "database",
        "start_period",
        "end_period",
        "action",
        "comment",
        "username",
    ]
    data_val = data_val[[c for c in data_val_cols if c in data_val.columns]]
    final_trends_cols = [
        c
        for c in [
            "well_id",
            "unit_hyphen",
            "aquifer",
            "slope_yr",
            "status_change",
            "status",
            "status_combined",
            "n_trend_obs",
        ]
        if c in final_trends.columns
    ]

    bcmap = BoMClassesColormap

    # Summary and details table for Current WL ranks
    curr_ranks["class_index"] = curr_ranks["rswl_bom_class"].map(
        {c: bcmap.class_names.index(c) for c in bcmap.class_names}
    )
    curr_ranks = curr_ranks
    curr_ranks_ct = curr_ranks["rswl_bom_class"].value_counts().fillna(0)
    curr_ranks_summ = curr_ranks_ct.to_frame(name="n_wells").fillna(0)
    curr_ranks_summ["pct_wells"] = (curr_ranks_ct / curr_ranks_ct.sum() * 100).round(2)
    curr_ranks_summ = curr_ranks_summ.reindex(bcmap.class_names).fillna(0)

    crs = curr_ranks_summ
    curr_ranks_summ_grpd = {
        "'Below average' or lower": crs.loc[
            ["Below average", "Very much below average", "Lowest on record"]
        ].sum(axis=0),
        "Average": crs.loc["Average"],
        "'Above average' or higher": crs.loc[
            ["Above average", "Very much above average", "Highest on record"]
        ].sum(axis=0),
    }
    curr_ranks_summ_grpd = pd.DataFrame(curr_ranks_summ_grpd).fillna(0).T

    crf_cols = [
        "rswl_bom_class",
        "well_id",
        "unit_hyphen",
        "aquifer",
        "class_index",
        "rswl",
        "obs_date",
    ]
    logger.debug(f"curr_ranks.columns = {curr_ranks.columns.values}")
    curr_ranks = curr_ranks[crf_cols]
    logger.debug(f"curr_ranks.columns = {curr_ranks.columns.values}")
    curr_ranks = curr_ranks.sort_values(["class_index"])

    wltrends_summ_by_triclass = (
        final_trends.groupby(["status_change"])
        .slope_yr.describe()
        .round(2)
        .rename(columns={"count": "n_wells"})
    )
    if wltrends_summ_by_triclass.n_wells.sum() > 0:
        wltrends_summ_by_triclass.insert(
            1,
            "pct_wells",
            (
                wltrends_summ_by_triclass.n_wells
                / wltrends_summ_by_triclass.n_wells.sum()
                * 100
            ).round(2),
        )
    wltrends_summ_all = final_trends[["slope_yr"]].describe().round(2).T
    final_trends = final_trends.sort_values(["slope_yr"])

    if len(total_wl_changes):
        total_wl_changes_summ_1 = (
            total_wl_changes.groupby(["window"]).change.describe().round(2)
        )
        total_wl_changes_summ_2 = (
            total_wl_changes.groupby(["window", "change_direction"])
            .change.describe()
            .round(2)
        )
    else:
        total_wl_changes_summ_1 = pd.DataFrame()
        total_wl_changes_summ_2 = pd.DataFrame()

    total_wl_changes = total_wl_changes.sort_values(
        ["window", "change_direction", "change"]
    ).set_index(["window", "change_direction"])

    retdata = {
        "unlisted_wells": sorted(unlisted_wells),
        "curr_rank_wells": sorted(curr_rank_wells),
        "curr_rank_excluded_wells": sorted(curr_rank_excluded_wells),
        "final_trend_wells": sorted(final_trend_wells),
        "excluded_trend_wells": sorted(excluded_trend_wells),
        "resource": resource,
        "wells": wells.to_dict(orient="records"),
        "wells_html": wells.to_html(escape=False),
        "data_val": data_val,
        "curr_ranks": curr_ranks[crf_cols],
        "curr_ranks_summ": curr_ranks_summ,
        "curr_ranks_summ_grpd": curr_ranks_summ_grpd,
        "majority_grpd_curr_ranks": curr_ranks_summ_grpd.sort_values("n_wells").iloc[
            -1
        ],
        "curr_ranks_qc": curr_ranks_qc,
        "final_trends": final_trends[final_trends_cols],
        "wltrends_summ_by_triclass": wltrends_summ_by_triclass,
        "wltrends_summ_all": wltrends_summ_all,
        "total_wl_changes": total_wl_changes,
        "total_wl_changes_summ_1": total_wl_changes_summ_1,
        "total_wl_changes_summ_2": total_wl_changes_summ_2,
    }

    return retdata


def construct_waterlevel_template_sentences(
    resource, data, highlight_method="papayawhip"
):
    """Construct template sentences for water level sections.

    Args:
        resource (wrap_technote.Resource): resource to summaries
        data (dict): collated water level results - see
            :func:`wrap_technote.collate_waterlevel_summary_data`
        highlight_method (str): "papayawhip" by default -
            see :func:`wrap_technote.utils.highlight_fields` for
            details

    Returns:
        list: list of sentences

    """
    maj = data["majority_grpd_curr_ranks"]
    total_wells = len(data["curr_ranks"])

    sentence_1 = (
        f"In <|{resource.trend_dfn.end_recovery_season:.0f}|>, "
        f"winter recovered water levels in "
        f"<|{maj.n_wells:.0f}|> out of <|{total_wells:.0f}|> (<|{maj.pct_wells:.0f}|>%) "
        "of monitoring wells in the "
        f"<|{resource.report_resources_mapping.tech_note_sentence_value}|> "
        f"are classified as <|{maj.name}|> (Section XYZ)."
    )

    tchanges = data["total_wl_changes_summ_1"]
    word = lambda x: "decline" if x < 0 else "rise"

    sentences_2 = []
    for year_range, row in tchanges.sort_index(ascending=False).iterrows():
        n_years = int(year_range.split()[0])
        if n_years in decimal_words.index.values:
            year_range_word = decimal_words[n_years]
        else:
            year_range_word = n_years
        sentences_2.append(
            (
                f"Over the past <|{year_range_word}|> years, "
                f"variations in water level in <|{row['count']:.0f}|> well<|{'s' if row['count'] > 1 else ''}|> "
                f"range from a <|{word(row['min'])}|> of <|{abs(row['min']):.2f}|> m "
                f"to a <|{word(row['max'])}|> of <|{abs(row['max']):.2f}|> m "
                f"(median is a <|{word(row['50%'])}|> of <|{abs(row['50%']):.1f}|> m)."
            )
        )

    trends = data["wltrends_summ_by_triclass"]
    majtr = get_majority_categories(trends)
    trend_years = resource.trend_dfn.end_year - resource.trend_dfn.start_year + 1
    words = ["two", "three", "four", "five", "six", "seven", "eight", "nine", "ten"]
    num2word = {i + 2: words[i] for i in range(len(words))}
    years_word = num2word[trend_years]
    years_word = years_word[0].upper() + years_word[1:]

    if word(majtr["min"]) == word(majtr["max"]):
        sentence_3 = (
            f"<|{years_word}|>-year trends show <|{majtr.name}|> water levels "
            f"in the majority of wells (<|{majtr.pct_wells:.0f}|>%), "
            f"with rates of <|{word(majtr['min'])}|> ranging between "
            f"<|{abs(majtr['min']):.2f}|>-<|{abs(majtr['max']):.2f}|> m/y "
            f"(median is a <|{word(majtr['50%'])}|> of <|{abs(majtr['50%']):.2f}|> m/y)."
        )
    else:
        sentence_3 = (
            f"<|{years_word}|>-year trends show <|{majtr.name}|> water levels "
            f"in the majority of wells (<|{majtr.pct_wells:.0f}|>%), "
            f"with rates ranging from a "
            f"<|{word(majtr['min'])}|> of <|{abs(majtr['min']):.2f}|> m/y to a "
            f"<|{word(majtr['max'])}|> of <|{abs(majtr['max']):.2f}|> m/y "
            f"(median <|{word(majtr['50%'])}|> of <|{abs(majtr['50%']):.2f}|> m/y)."
        )

    sentences = [sentence_1] + sentences_2 + [sentence_3]

    sentences = highlight_fields(sentences, highlight_method=highlight_method)

    return sentences


class QueryAndValidateWaterLevels(Run):
    """First stage of water level analysis. This downloads water level
    data (both manual and logger), filters it, and creates data validation
    charts.

    Args:
        resource (:class:`wrap_technote.Resource`)
        show_comments (bool): whether to show SA Geodata comments on the
            for the manual water levels, on the data validation charts
        plot_figures (bool): create the data validation figures

    Attributes:
        data (dict): data artifacts created during the stage

    The steps of this stage are:

    1. :meth:`wrap_technote.QueryAndValidateWaterLevels.load_resource_definition`
    2. :meth:`wrap_technote.QueryAndValidateWaterLevels.fetch_wl_data`
    3. :meth:`wrap_technote.QueryAndValidateWaterLevels.drop_null_data`
    4. :meth:`wrap_technote.QueryAndValidateWaterLevels.filter_wl_data`
    5. :meth:`wrap_technote.QueryAndValidateWaterLevels.add_well_titles`
    6. :meth:`wrap_technote.QueryAndValidateWaterLevels.make_wl_validation_charts`
       (only run if plot_figures is True, obviously!)
    7. :meth:`wrap_technote.QueryAndValidateWaterLevels.save_results`

    """

    def __init__(
        self,
        resource,
        show_comments: bool = True,
        plot_figures: bool = True,
    ):
        super().__init__(
            resource=resource, show_comments=show_comments, plot_figures=plot_figures
        )

        for func in [
            self.load_resource_definition,
            self.fetch_wl_data,
            self.drop_null_data,
            self.filter_wl_data,
            self.add_well_titles,
            self.if_then_use("plot_figures", self.make_wl_validation_charts),
            self.save_results,
        ]:
            self.append_step(func)

    def load_resource_definition(self, resource, **kwargs):
        parameter = resource.resource_key.split("_")[-1]
        wells = resource.find_wells()
        return {"wells": wells}

    def fetch_wl_data(self, wells: pd.DataFrame, **kwargs):
        wls = gd.fetch_wl_data(wells, include_replacements=False)
        return {"wls": wls}

    def drop_null_data(self, wls: pd.DataFrame, **kwargs):
        wls = wls.dropna(subset=["dtw", "swl", "rswl"], how="all")
        return {"wls": wls}

    def filter_wl_data(self, wls: pd.DataFrame, resource, **kwargs):
        wls, removals = filter_wl_observations(wls, return_removals=True, qc=resource)
        wls = wls.groupby("well_id").filter(lambda x: len(x) >= 5)
        return {"wls": wls, "removals": removals}

    def add_well_titles(self, wls: pd.DataFrame, removals: pd.DataFrame, **kwargs):
        for df in (wls, removals):
            df["well_title"] = df.unit_hyphen + " " + df.obs_no + " " + df.dh_name
        return {"wls": wls, "removals": removals}

    def make_wl_validation_charts(
        self,
        wls: pd.DataFrame,
        removals: pd.DataFrame,
        resource,
        show_comments: bool,
        **kwargs,
    ):
        axes = plot_wl_data_validation(
            wls,
            removals,
            well_title_col="well_title",
            show_comments=show_comments,
            adjust_comments=False,
            path=resource.static_path,
            savefig_and_close=True,
            dpi=100,
        )

    def save_results(
        self, resource, wls: pd.DataFrame, removals: pd.DataFrame, **kwargs
    ):
        table_prefix = "validated_data"
        output_data = {"valid_data": wls, "invalid_data": removals}
        resource.write_data_sheets(table_prefix, output_data)
        output_data = {table_prefix + "__" + k: v for k, v in output_data.items()}
        return {"output_data": output_data}


class CalculateSeasonalWaterLevels(Run):
    """Second stage of water level analysis. This calculates an annual
    'recovered' water level for each well, for each year.

    Args:
        resource (:class:`wrap_technote.Resource`)
        plot_figures (bool): create the data validation figures

    Attributes:
        data (dict): data artifacts created during the stage

    The steps of this stage are:

    1. :meth:`wrap_technote.CalculateSeasonalWaterLevels.load_data`
    2. :meth:`wrap_technote.CalculateSeasonalWaterLevels.calculate_seasonal_water_levels`
    3. :meth:`wrap_technote.CalculateSeasonalWaterLevels.filter_to_recovered_wls_only`
    4. :meth:`wrap_technote.CalculateSeasonalWaterLevels.calculate_quality_of_record`
    5. :meth:`wrap_technote.CalculateSeasonalWaterLevels.apply_data_validations`
    6. :meth:`wrap_technote.CalculateSeasonalWaterLevels.rank_and_classify_recovered_wls`
    7. :meth:`wrap_technote.CalculateSeasonalWaterLevels.clean_up`
    8. :meth:`wrap_technote.CalculateSeasonalWaterLevels.calculate_total_waterlevel_changes`
    9. :meth:`wrap_technote.CalculateSeasonalWaterLevels.make_charts`
    10. :meth:`wrap_technote.CalculateSeasonalWaterLevels.save_data`

    """

    def __init__(
        self,
        resource,
        plot_figures: bool = True,
    ):
        super().__init__(resource=resource, plot_figures=plot_figures)

        for func in [
            self.load_data,
            self.calculate_seasonal_water_levels,
            self.filter_to_recovered_wls_only,
            self.calculate_quality_of_record,
            self.apply_data_validations,
            self.rank_and_classify_recovered_wls,
            self.clean_up,
            self.calculate_total_waterlevel_changes,
            self.if_then_use("plot_figures", self.make_charts),
            self.save_data,
        ]:
            self.append_step(func)

    def load_data(self, resource, **kwargs) -> dict:
        wls = resource.read_data("validated_data__valid_data")
        return {"wls": wls}

    def calculate_seasonal_water_levels(self, wls, resource, **kwargs) -> dict:
        func = apply_calculate_seasonal_water_levels(resource)
        wlann = wls.groupby("well_id").apply(func).reset_index(drop=True)
        return {"wlann": wlann}

    def filter_to_recovered_wls_only(self, wlann, resource, **kwargs):
        # Create a subset of rows with only the recovery water levels.
        wlrec = wlann.loc[wlann.season == "recovery", :]

        # Remove any recovered levels which are beyond end_recovery_season
        end_y = int(str(resource.trend_dfn.end_recovery_season)[:4])
        wlrec = wlrec[wlrec.season_year.astype(str).str[:4].astype(int) <= end_y]
        return {"wlrec": wlrec}

    def calculate_quality_of_record(self, wlrec, resource, **kwargs):
        qc, cmp_df = resource.get_wl_ranking_qc_results(wlrec)
        logger.debug(
            str(
                pd.merge(
                    qc, cmp_df["all_conditions"], left_index=True, right_index=True
                )
            )
        )
        return {"qc": qc, "cmp_df": cmp_df}

    def apply_data_validations(self, resource, cmp_df, qc, **kwargs):
        vals = resource.read_table("Data_validation")
        incl = vals.query("action == 'Include well in water level rankings'")
        for ix, row in incl.iterrows():
            for well_id in row.well_id.split(","):
                well_id = well_id.strip()
                if well_id in cmp_df.index:
                    x = cmp_df.loc[well_id, "all_conditions"]
                    cmp_df.loc[well_id, "all_conditions"] = True
                    cmp_df.loc[
                        well_id, "comment"
                    ] = f"Included True by {row.username}: {row.comment} (overriding {x})"
        data_quality = pd.merge(
            qc, cmp_df, left_index=True, right_index=True
        ).reset_index()
        return {
            "vals": vals,
            "incl": incl,
            "cmp_df": cmp_df,
            "data_quality": data_quality,
        }

    def rank_and_classify_recovered_wls(self, wlrec, resource, **kwargs):
        wlranks = pd.merge(
            wlrec,
            wlrec.groupby("well_id", sort=False).rswl.apply(
                lambda df: tn.rank_and_classify(df, "rswl")
            ),
            left_index=True,
            right_index=True,
        )
        end_year = int(resource.trend_dfn.end_year)
        next_year = end_year + 1
        next_year2 = str(int(next_year))[2:]
        end_year2 = f"{end_year}-{next_year2}"
        curr_year_recovery_seasons = [f"{end_year}-recovery", f"{end_year2}-recovery"]
        logger.info(f"Current year recovery seasons: {curr_year_recovery_seasons}")
        wlranks_curr = wlranks[wlranks["year+season"].isin(curr_year_recovery_seasons)]
        return {"wlranks": wlranks, "wlranks_curr": wlranks_curr, "end_year": end_year}

    def clean_up(self, wlrec, wlranks, data_quality, wlranks_curr, **kwargs):
        wlrec = wlrec.reset_index()
        wlranks = (
            pd.merge(wlranks, data_quality[["well_id", "all_conditions"]], on="well_id")
            .rename(columns={"all_conditions": "meets_qc_filter"})
            .reset_index()
        )
        wlranks_curr = (
            pd.merge(
                wlranks_curr, data_quality[["well_id", "all_conditions"]], on="well_id"
            )
            .rename(columns={"all_conditions": "meets_qc_filter"})
            .reset_index()
        )

        wlranks_included = wlranks[wlranks.meets_qc_filter == True]
        wlranks_excluded = wlranks[wlranks.meets_qc_filter == False]
        wlranks_curr_included = wlranks_curr[wlranks_curr.meets_qc_filter == True]
        wlranks_curr_excluded = wlranks_curr[wlranks_curr.meets_qc_filter == False]
        return {
            "wlrec": wlrec,
            "wlranks": wlranks,
            "wlranks_curr": wlranks_curr,
            "wlranks_included": wlranks_included,
            "wlranks_excluded": wlranks_excluded,
            "wlranks_curr_included": wlranks_curr_included,
            "wlranks_curr_excluded": wlranks_curr_excluded,
        }

    def calculate_total_waterlevel_changes(self, wlranks, end_year, **kwargs):
        windows = get_total_wl_change_windows(current_year=end_year)
        total_wl_changes = get_total_wl_changes(wlranks, windows=windows).reset_index()
        return {"total_wl_changes": total_wl_changes}

    def make_charts(
        self,
        resource,
        wlranks,
        wls,
        cmp_df,
        wlann,
        wlranks_curr,
        data_quality,
        **kwargs,
    ):
        for fn in resource.static_path.glob("plot_wl_bom_classes_*.png"):
            logger.debug(f"removing old figure: {fn}")
            os.remove(fn)
        for fn in resource.static_path.glob("plot_wl_seasonality_*.png"):
            logger.debug(f"removing old figure: {fn}")
            os.remove(fn)

        logger.info(
            "Chart the original (filtered) water levels against the ranked and classified recovered water levels..."
        )
        for well_id, wwlranks in wlranks.groupby(wlranks.well_id):
            logger.info(f"{well_id}...")
            r = tn.plot_wl_bom_classes(
                wwlranks, wls[wls.well_id == well_id], fig=plt.figure(figsize=(10, 4))
            )
            titledf = wls[wls.well_id == well_id].iloc[0]
            well_title = f"{titledf.unit_hyphen} {titledf.obs_no} {titledf.dh_name}"
            r["ax"].set_title(
                f"{well_title}\nmeets all_conditions {cmp_df.loc[well_id].all_conditions}"
            )
            r["ax"].figure.tight_layout()
            filename = str(resource.static_path / f"plot_wl_bom_classes_{well_id}.png")
            logger.debug(f"filename = {filename}")
            r["ax"].figure.savefig(filename, dpi=100)
            tn.trim_whitespace_from_image(filename, append=False)

        logger.info("Chart seasonality for each well...")
        for well_id, wwlann in wlann.groupby("well_id"):
            logger.info(f"Plotting seasonal chart for {well_id}")
            seasons = resource.get_season(well_id)
            r = tn.plot_wl_seasonality(
                wls[wls.well_id == well_id], wwlann, seasons=seasons, dots=True
            )
            titledf = wls[wls.well_id == well_id].iloc[0]
            well_title = f"{titledf.unit_hyphen} {titledf.obs_no} {titledf.dh_name}"
            r["fig"].suptitle(well_title)
            filename = str(resource.static_path / f"plot_wl_seasonality_{well_id}.png")
            r["fig"].savefig(filename, dpi=100)
            tn.trim_whitespace_from_image(filename, append=False)

        logger.info("Charting: plot_wl_rankings_internal")
        fig = tn.plot_wl_rankings_internal(wlranks_curr, title=resource.resource_key)
        filename = f"{resource.resource_key}_wl_rankings_internal.png"
        for path in [resource.data_path]:
            fn = str(path / filename)
            fig.savefig(fn, dpi=120)
            tn.trim_whitespace_from_image(fn, append=False)

        logger.info("Charting: plot_wl_historical_rankings")
        fig = tn.plot_wl_historical_rankings(
            wlranks, data_quality, title=resource.resource_key
        )
        filename = f"{resource.resource_key}_wl_historical_rankings.png"
        for path in [resource.data_path]:
            fn = str(path / filename)
            fig.savefig(fn, dpi=120)
            tn.trim_whitespace_from_image(fn, append=False)

    def save_data(
        self,
        resource,
        data_quality,
        wlrec,
        wlranks_curr,
        wlranks_included,
        wlranks_curr_included,
        wlranks_excluded,
        wlranks_curr_excluded,
        total_wl_changes,
        wlranks,
        wlann,
        **kwargs,
    ):
        data_quality = data_quality.reset_index()

        output_data = {
            "annual_season_wl": wlann,
            "all_ranked_wl": wlranks,
            "annual_recovered_wl": wlrec,
            "ranked_wls": wlranks_included,
            "current_ranked_wls": wlranks_curr_included,
            "ranks_excl": wlranks_excluded,
            "current_ranks_excl": wlranks_curr_excluded,
            "total_wl_changes": total_wl_changes,
            "data_quality": data_quality,
        }
        table_prefix = "recovery_wl_data"
        resource.write_data_sheets("recovery_wl_data", output_data)

        for path in [resource.data_path]:
            spath = path / "spatial_outputs"
            spath.mkdir(exist_ok=True)
            tn.df_to_shp(
                wlranks, spath / f"ranked_recovered_wls_{resource.resource_key}.shp"
            )
            tn.df_to_shp(
                wlranks_curr,
                spath / f"ranked_current_recovered_wls_{resource.resource_key}.shp",
            )

        output_data = {table_prefix + "__" + k: v for k, v in output_data.items()}
        return {"output_data": output_data}


waterlevel_analysis_stages = [
    QueryAndValidateWaterLevels,
    CalculateSeasonalWaterLevels,
]



import colorsys
from datetime import date, datetime, timedelta
import io
import re
import os
import logging
import textwrap
from pathlib import Path
import copy

import matplotlib.pyplot as plt
from matplotlib import lines as mlines
from matplotlib import gridspec
from matplotlib import dates as mdates
from matplotlib import ticker as mticker
from matplotlib import colors as mcolors
from matplotlib import patches as mpatches
import numpy as np
import pandas as pd
from adjustText import adjust_text
from scipy import stats
from PIL import Image, ImageChops
import pyproj

from .gwutils import *
from .utils import *
from .charts_utils import *

logger = get_logger()


def plot_tds_data_validation(
    df,
    removals=None,
    show_comments=True,
    well_id_col="well_id",
    dt_col="collected_date",
    tds_col="tds",
    label_col="reason",
    mew=0.3,
    adjust_comments=False,
    adjust_text_lim=30,
    ignore_comment_regexps=(r"Hydat.*", r".*LIC ?= ?.*"),
    savefig_and_close=False,
    path=".",
    dpi=130,
    fig=None,
    **kws,
):
    """Plot a graph of salinity data for validation purposes.

    Args:
        df (pandas DataFrame): salinity data for one or more wells
        removals (pandas DataFrame): salinity data which has been deemed invalid
            but should be included on the plot (to see what has been removed and
            why), for one or more wells.
        show_comments (bool): show comments on the chart
        well_id_col (str): column label for a unique well identifier
        dt_col (str): column label for the observation date in datetimes
        tds_col (str): column label with salinity data
        label_col (str): column label for *removals* with the reason for the
            data points removal
        mew (float): width of line
        adjust_comments (bool): use the adjustText python library to attempt
            and move the comment labels so that they do not overlap. It is quite
            slow, so be aware of that.
        adjust_text_lim (int): number of iterations for adjustText to run. 2 or 3
            will be fast but likely ineffective for all but the simplest plots.
            100 will be quite slow but a thorough test on almost all plots. 30
            is a good compromise.
        ignore_comment_regexps (sequence of str): each string is a regexp pattern.
            If it matches the comment for a given data point, the comment will not
            be plotted on the chart (only relevant if show_comments is True).
        path (str): path to save figures into.
        savefig_and_close (bool): save fig to png and close
        dpi (int): resolution of saved PNGs if *savefig_and_close=True*.

    Any keyword arguments beginning with ``adjust_text`` will be sent to the
    :func:`adjustText.adjust_text` function, so for example to tell adjustText to
    use autoalign, pass *adjust_text_autoalign=True* to this function.

    Returns: a list of matplotlib Axes.

    """
    if removals is not None:
        all_removal_well_ids = list(removals[well_id_col].unique())
    else:
        all_removal_well_ids = []

    well_ids = sorted(list(set(list(df[well_id_col].unique()) + all_removal_well_ids)))

    df.loc[df.extract_method.isnull(), "extract_method"] = "?"
    df.loc[df.measured_during.isnull(), "measured_during"] = "?"

    comment_props = dict(marker="o", lw=0, mfc="none", mec="purple", mew=0.3, ms=12)

    def add_comments(cdf, cax):
        ctexts = []
        if len(cdf):
            cax.plot(cdf[dt_col], cdf[tds_col], **comment_props)
            for idx, row in cdf.iterrows():
                comment = str(row.comments)
                x = row[dt_col]
                y = row[tds_col]
                by = str(row.collected_by).strip()
                by = "" if by == "None" else by
                if by:
                    comment = f"{by}: {comment}"
                if comment and isinstance(x, datetime) and not np.isnan(y):
                    add_label = True
                    for fail_pattern in ignore_comment_regexps:
                        if re.match(fail_pattern, comment):
                            add_label = False
                    if add_label:
                        text = cax.text(x, y, wrap(comment), color="purple", fontsize=5)
                        ctexts.append(text)
        return ctexts

    axes = []
    if savefig_and_close:
        if not fig:
            fig = plt.figure(figsize=(10, 4))

    for i, well_id in enumerate(well_ids):
        logger.debug(f"Charting {well_id}")
        if savefig_and_close:
            ax = fig.add_subplot(111)
        else:
            fig = plt.figure(figsize=(7, 3))
            ax = fig.add_subplot(111)

        ax.set_title(well_id)

        w_df = df[df[well_id_col] == well_id]

        texts = []

        # Group the salinity data by extract_method and measured_during
        for extract_method, wdf2 in w_df.groupby("extract_method"):
            for mduring, wdf3 in wdf2.groupby("measured_during"):
                # Chart this group of salinity data with
                # the marker shape representing the extract_method; and
                # the fill colour representing the measured_during code
                ax.plot(
                    wdf3[dt_col],
                    wdf3[tds_col],
                    marker=extract_method_markers[extract_method],
                    mfc=meas_during_colours[mduring],
                    mec="black",
                    mew=mew,
                    lw=0,
                    label="",
                )
                if show_comments:
                    texts += add_comments(wdf3[~wdf3.comments.isnull()], ax)

        # Chart a line connecting all the salinity data points.
        (sal_line,) = ax.plot(
            w_df[dt_col],
            w_df[tds_col],
            ls="-",
            lw=0.5,
            color="tab:cyan",
            label="All retained data",
        )

        if removals is not None:
            w_removals = removals[removals[well_id_col] == well_id]
            removal_reasons = list(w_removals.reason.unique())

            # Chart the values which have been excluded (and the reason why).
            for j, (reason, rdf) in enumerate(w_removals.groupby("reason")):
                ax.plot(
                    rdf[dt_col],
                    rdf[tds_col],
                    marker="X",
                    color=removal_reason_colours[removal_reasons.index(reason)],
                    lw=0,
                    label=reason,
                )
            if show_comments:
                texts += add_comments(w_removals[~w_removals.comments.isnull()], ax)

        # Add a legend item for comments
        ax.plot([], [], label="SA Geodata comments", **comment_props)

        # Add a legend item for the differnt extract_method codes.
        for emeth in w_df.extract_method.unique():
            ax.plot(
                [],
                [],
                lw=0,
                marker=extract_method_markers[emeth],
                color="k",
                mew=mew,
                mfc="white",
                label=f"Extract by: {EXTRACT_METHOD_LUT[emeth]}",
            )

        # Add a legend item for the different measured_during codes.
        for mduring in w_df.measured_during.unique():
            ax.plot(
                [],
                [],
                lw=0,
                marker="8",
                mew=0,
                mfc=meas_during_colours[mduring],
                label=f"Meas. during {measured_during_lut[mduring]}",
            )

        # Show ignore_comment_patterns in the legend.
        if show_comments and ignore_comment_regexps:
            for fail_pattern in ignore_comment_regexps:
                plt.plot(
                    [],
                    [],
                    ls="",
                    label=f'(Comments matching regexp\n"{fail_pattern}" not shown)',
                )

        # Add the legend
        ax.legend(loc="best", fontsize="xx-small", frameon=False, ncol=2)

        # Rearrange the labels if requested
        if adjust_comments:
            logger.debug(
                f"Using adjust_text to shift comment annotations for {well_id}"
            )
            adjust_text_kws = {}
            adjust_text_kws["arrowprops"] = kws.get(
                "adjust_text_arrowprops", dict(arrowstyle="->", color="gray", lw=0.3)
            )
            adjust_text_kws["autoalign"] = kws.get("adjust_text_autoalign", False)
            adjust_text_kws["lim"] = adjust_text_lim
            for key, argval in kws.items():
                if key.startswith("adjust_text"):
                    adjust_text_kws[key.replace("adjust_text_", "")] = argval
            adjust_text(
                texts,
                x=mdates.date2num(sal_line.get_xdata()),
                y=sal_line.get_ydata(),
                **adjust_text_kws,
            )
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.set_ylabel(tds_col)
        ax.set_xlabel(dt_col)

        if savefig_and_close:
            fig.savefig(str(path / f"plot_tds_data_validation_{well_id}.png"), dpi=dpi)
            fig.clf()
        else:
            axes.append(ax)

    df.loc[df.extract_method == "?", "extract_method"] = None
    df.loc[df.measured_during == "?", "measured_during"] = None

    return axes


def plot_salinity_trend(
    sals,
    trend_sals,
    trend_lines=None,
    excluded_sals=None,
    well_id_col="well_id",
    dt_col="collected_date",
    sal_col="tds",
    hist_max_date_col="tds_max_date",
    ax=None,
    ms=5,
    year_span=12,
    override_year_span=False,
):
    well_id = sals[well_id_col].unique()[0]
    logger.debug(f"Charting {well_id}")
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    lw = 0.5
    ax.plot(
        sals[dt_col],
        sals[sal_col],
        color="#222222",
        lw=lw,
        marker="o",
        mfc="orange",
        mec="gray",
        ms=ms - 1,
        mew=0.3,
        alpha=0.5,
        label="All valid salinity data",
    )
    if excluded_sals is not None:
        ax.plot(
            excluded_sals[dt_col],
            excluded_sals[sal_col],
            lw=0,
            marker="o",
            ms=ms,
            mew=0.2,
            mec="grey",
            mfc="grey",
            color="grey",
            alpha=0.2,
            label="Excluded data",
        )
    ax.plot(
        trend_sals[dt_col],
        trend_sals[sal_col],
        marker="o",
        mfc="red",
        mec="k",
        ms=ms,
        mew=0.3,
        lw=0,
        alpha=0.7,
        label="Used for trend calculation",
    )
    hist_idx = sals[dt_col] == trend_sals[hist_max_date_col].iloc[0]
    ax.plot(
        sals[hist_idx][dt_col],
        sals[hist_idx][sal_col],
        marker="o",
        mec="purple",
        mfc="none",
        mew=0.5,
        ms=ms + 5,
        ls="none",
        label="Historical maximum",
    )

    if trend_lines is not None:
        start_year = min(trend_sals[dt_col]).year
        end_year = max(trend_sals[dt_col]).year
        trend_line = trend_lines.loc[well_id]
        trend_line_ts = [
            pd.Timestamp(f"{start_year}-01-01"),
            pd.Timestamp(f"{end_year}-12-31"),
        ]
        trend_line_vals = calculate_trendline_at_dates(trend_line, trend_line_ts)
        ax.plot(trend_line_ts, trend_line_vals, lw=1, color="r", label="Linear trend")
        end_ts = trend_line_ts[1]
        fg, bg = status_to_colours(trend_line.status_change, param="TDS")
        trend_label = (
            f"{well_id}: {trend_line.status_change}"
            f" at {trend_line.slope_yr:.0f} mg/L/y ({trend_line.sal_pct_change:.1f}%)"
        )
    else:
        end_ts = pd.Timestamp(trend_sals[dt_col].max())
        fg = "black"
        bg = "white"

    start_ts = end_ts - timedelta(days=365.25 * year_span)
    if "max" in str(override_year_span):
        hist_max_ts = sals[hist_idx][dt_col].iloc[0] - timedelta(days=120)
        if hist_max_ts < start_ts:
            start_ts = hist_max_ts
    elif override_year_span == "all":
        start_ts = sals[dt_col].min() - timedelta(days=120)

    ax.set_xlim(start_ts, end_ts + timedelta(days=120))
    leg = ax.legend(loc="best", fontsize="x-small", frameon=False)
    ax.set_ylabel(sal_col)
    title_text = ax.set_title(
        trend_label, color=fg, backgroundcolor=bg, fontsize="medium"
    )

    lim_sals = sals[(sals[dt_col] >= start_ts) & (sals[dt_col] <= end_ts)][sal_col]
    yspan = lim_sals.max() - lim_sals.min()
    y0 = lim_sals.min() - yspan * 0.1
    y1 = lim_sals.max() + yspan * 0.1
    ax.set_ylim(y0, y1)
    ax._title_text = title_text
    return ax


def plot_tds_trends_internal(tdstrends, title="", param="mg"):
    """param can be 'mg' or '5_yr_pct'"""
    if param == "mg":
        param_col = "slope_yr"
        yaxis_label = "mg/L/y"
    elif param == "5yr_pct":
        param_col = "sal_pct_change"
        yaxis_label = "% change over 5 yrs"
    assert param in ("mg", "5yr_pct")
    return plot_trends_internal(
        tdstrends,
        title=title,
        class_names=("Decreasing", "Stable", "Increasing"),
        param_yaxis_label=yaxis_label,
        param_col=param_col,
    )


def plot_tds_current_internal(df, title="", ax=None, fig=None):
    anntds = df.copy()
    anntds["tds_interval"] = pd.cut(anntds.tds, np.arange(0, 10000, 200))
    tdsbins = anntds.groupby("tds_interval").dh_no.count().to_frame().reset_index()
    tdsbins["left"] = [interval.left for interval in tdsbins.tds_interval]
    tdsbins["right"] = [interval.right for interval in tdsbins.tds_interval]
    tdsbins["centre"] = tdsbins.left + (tdsbins.right - tdsbins.left) / 2

    if fig is None:
        fig = plt.figure(figsize=(8, 3))
        ax = fig.add_subplot(111)

    highest_tds = 0
    label = "first"
    for ix, row in tdsbins.iterrows():
        if row.dh_no > 0:
            if label == "first":
                label = "Number of wells\n(200 mg/L bins)"
            else:
                label = ""
            ax.bar(
                [row.centre],
                [row.dh_no],
                color="tab:blue",
                alpha=1,
                width=150,
                label=label,
            )
            highest_tds = row.right
    median_tds = anntds.tds.median()
    mean_tds = anntds.tds.mean()
    ax.axvline(
        median_tds, lw=2, color="tab:red", label=f"median ({median_tds:.0f} mg/L)"
    )
    ax.axvline(
        mean_tds,
        lw=2,
        color="purple",
        dashes=(5, 1),
        label=f"mean ({mean_tds:.0f} mg/L)",
    )
    ax.legend(loc="best", fontsize="small", frameon=False)
    ax.xaxis.set_major_locator(mticker.MaxNLocator(15))
    ax.set_xlim(0, highest_tds)
    ax.set_ylabel("Number of wells")
    ax.set_xlabel("Current salinity (mg/L)")
    ax.grid(True, lw=0.5, color="k", alpha=0.5, ls=":")
    ax.grid(True, axis="x", lw=0.5, color="k", alpha=0.5, ls=":")
    ax.set_title(title)
    return ax


def plot_tds_resource_monitoring_history(annstats, axes=None, fig=None):
    if axes is None:
        fig = plt.figure(figsize=(8, 6))
        gs = gridspec.GridSpec(2, 1, height_ratios=(4, 2), hspace=0)
        ax = fig.add_subplot(gs[0])
        ax2 = fig.add_subplot(gs[1], sharex=ax)
    else:
        ax, ax2 = axes

    lower = annstats.tds_mean - annstats.tds_stdev
    lower2 = (lower - annstats.tds_stdev).dropna()
    upper = annstats.tds_mean + annstats.tds_stdev
    upper2 = (upper + annstats.tds_stdev).dropna()

    logger.debug(f"lower2: {lower2}")
    logger.debug(f"upper2: {upper2}")

    if len(lower):
        ax.fill_between(
            annstats.index,
            lower,
            upper,
            color="tab:blue",
            alpha=0.8,
            lw=0,
            label="±1 s.d.",
        )
        if len(lower2):
            if len(lower) == len(lower2):
                ax.fill_between(
                    annstats.index,
                    lower2,
                    lower,
                    color="lightblue",
                    lw=0,
                    label="±2 s.d.",
                )
        if len(upper2):
            if len(upper) == len(upper2):
                ax.fill_between(
                    annstats.index, upper, upper2, color="lightblue", lw=0, label=""
                )

    ax.plot(annstats.index, annstats.tds_mean, color="darkblue", lw=1, label="Mean TDS")
    ax.plot(
        annstats.index, annstats.tds_median, color="darkgreen", lw=1, label="Median TDS"
    )
    ax2.bar(
        annstats.index,
        annstats.n_wells,
        color="gray",
        lw=1,
        alpha=1,
        label="Number of wells",
    )

    ax.set_ylabel("Salinity (mg/L)")
    ax2.set_ylabel("Number of wells")

    ax.yaxis.set_major_locator(mticker.MaxNLocator(10))
    ax2.yaxis.set_major_locator(mticker.MaxNLocator(5))

    ax.legend(loc="center left", ncol=1, fontsize="small", frameon=True)
    ax2.legend(loc="upper left", ncol=4, fontsize="small", frameon=True)

    if len(annstats):
        ax2.set_ylim(0, annstats.n_wells.max())
        if len(lower2) and len(upper2):
            ax.set_ylim(max([lower2.min(), 0]), upper2.max())

    ax.grid(True, ls="--", lw=0.5)
    ax2.grid(True, ls="--", lw=0.5)

    ax.figure.tight_layout()
    return ax, ax2


def plot_tds_resource_variability(annstats, axes=None):
    if axes is None:
        fig = plt.figure(figsize=(9, 4))
        gs = gridspec.GridSpec(1, 2)
        ax1 = fig.add_subplot(gs[0])
        ax2 = fig.add_subplot(gs[1])
        axes = (ax1, ax2)
    for i, key in enumerate(("n_wells", "year")):
        if key == "n_wells":
            s = annstats.n_wells
            cmap = plt.cm.magma_r
        elif key == "year":
            s = annstats.index
            cmap = plt.cm.viridis
        ax = axes[i]
        cset = ax.scatter(
            annstats.tds_mean,
            annstats.tds_stdev / annstats.tds_mean * 100,
            c=s,
            marker="o",
            s=80,
            linewidths=0.5,
            edgecolors="k",
            cmap=cmap,
        )
        cbar = plt.colorbar(cset, ax=ax)
        cbar.set_label(key)
        ax.set_ylabel("Standard deviation (% of mean)")
        ax.set_xlabel("Mean TDS (mg/L)")
    axes[0].figure.tight_layout()
    return axes


# def plot_salinity_status_and_trend(df, tdf, results, axes=None):
#     """Plot salinity status and trend results.

#     Args:
#         df (pd.DataFrame): containing data
#         tdf (pd.DataFrame): containing the trend line data
#         results (pd.Series): containing the results of the calculation.

#     The data for this figure comes from the function :func:`get_historical_salinity_trends`,
#     i.e. the last entry in the list of dataframes, and the last row of the
#     "results" dataframe.

#     """
#     if not axes:
#         fig = plt.figure()
#         fig, axes = plt.subplots(2, figsize=(8, 9))
#         # The top figure shows the data; bottom figure shows derived indicators.
#         data_ax = axes[0]
#         sal_ind_ax = axes[1]
#         sal_trend_ax = sal_ind_ax.twinx()
#     else:
#         data_ax, sal_ind_ax, sal_trend_ax = axes

#     results = results.dropna(how="any")

#     (data_line,) = data_ax.plot(
#         df.collected_date,
#         df.tds,
#         lw=1,
#         ls="-",
#         color="k",
#         marker="o",
#         ms=10,
#         mec="k",
#         mfc="none",
#     )
#     (mean_line,) = data_ax.plot(
#         df.collected_date,
#         df["mean"],
#         lw=2,
#         color="tab:blue",
#         dashes=(2, 0.8),
#         label="Mean salinity",
#     )
#     (trend_line,) = data_ax.plot(
#         tdf.timestamp,
#         tdf.predicted,
#         lw=2,
#         color="tab:red",
#         dashes=(10, 2),
#         label="Trend line",
#     )
#     data_label = data_ax.text(
#         0.5,
#         0.96,
#         results.iloc[-1].label,
#         transform=data_ax.transAxes,
#         ha="center",
#         va="top",
#     )
#     data_ax.grid(True, color="grey", lw=0.5, ls=":")
#     tds_range = df.tds.max() - df.tds.min()
#     data_ax.set_ylim(df.tds.min() - tds_range * 0.05, df.tds.max() + tds_range * 0.2)
#     data_ax.legend(frameon=False, fontsize="small")

#     (sal_ind_line,) = sal_ind_ax.plot(
#         pd.to_datetime(results.year.astype(str)),
#         results.current_pct_diff,
#         lw=1,
#         marker="o",
#         ms=10,
#         mfc="none",
#     )
#     (sal_trend_line,) = sal_trend_ax.plot(
#         pd.to_datetime(results.year.astype(str)),
#         results.slope_pct_change_yr * 10,
#         lw=1,
#         marker="d",
#         color="tab:red",
#         ms=10,
#         mfc="none",
#     )
#     sal_ind_ax.axhline(0, lw=2, color="tab:blue", dashes=(10, 2), alpha=0.5)
#     sal_trend_ax.axhline(0, lw=2, color="tab:red", dashes=(3, 0.8), alpha=0.5)
#     sal_ind_ax.grid(True, axis="y", color="tab:blue", lw=0.5, dashes=(10, 2), alpha=0.8)
#     sal_trend_ax.grid(
#         True, axis="y", color="tab:red", lw=0.5, dashes=(3, 0.8), alpha=0.8
#     )
#     sal_ind_ax.grid(True, axis="x", color="grey", lw=0.5, ls=":")

#     sal_ind_ax.set_ylim(results.current_pct_diff.min(), results.current_pct_diff.max())

#     plt.setp(sal_ind_ax.get_yticklabels(), color="tab:blue")
#     plt.setp(sal_trend_ax.get_yticklabels(), color="tab:red")

#     sal_ind_ax.set_ylabel("Salinity difference from mean (%)", color="tab:blue")
#     sal_trend_ax.set_ylabel("Salinity trend (%/decade)", color="tab:red")

#     x0, x1 = pd.Timestamp(str(df.collected_date.dt.year.min() - 1)), pd.Timestamp(
#         str(df.collected_date.dt.year.max() + 1)
#     )
#     data_ax.set_xlim(x0, x1)
#     sal_ind_ax.set_xlim(x0, x1)

#     return data_ax, sal_ind_ax, sal_trend_ax


def tweak_salinity_yaxis_limits(ax, bins):
    data_lims = ax.get_ylim()
    pos_bins = [b for b in bins if b > 0]
    neg_bins = [b for b in bins if b < 0]
    new_min = min([max(neg_bins), min(data_lims)])
    new_max = max([min(pos_bins), max(data_lims)])
    ax.set_yticks(bins)
    ax.set_ylim(new_min, new_max)


def plot_salinity_status_and_trend_simple(df, tdf, results, ax=None):
    """Plot salinity status and trend results.

    Args:
        df (pd.DataFrame): containing data
        tdf (pd.DataFrame): containing the trend line data
        results (pd.Series): containing the results of the calculation.
        ax (Axes): optional Axes object

    Returns: Axes object

    The data for this figure comes from the function :func:`get_historical_salinity_trends`,
    i.e. the last entry in the list of dataframes, and the last row of the
    "results" dataframe. It differs from the other method in that this only plots
    the data and trend, not the historical version of the trend.

    See also :func:`plot_salinity_status_and_trend_simple_with_indicators`

    """
    if not ax:
        fig = plt.figure(figsize=(6, 4))
        ax = fig.add_subplot(111)

    results = results.dropna(how="any")

    (data_line,) = ax.plot(
        df.collected_date,
        df.tds,
        lw=1,
        ls="-",
        color="k",
        marker="o",
        ms=10,
        mec="k",
        mfc="none",
    )
    (mean_line,) = ax.plot(
        df.collected_date,
        df["mean"],
        lw=2,
        color="tab:blue",
        dashes=(2, 0.8),
        label="Mean salinity",
    )
    (trend_line,) = ax.plot(
        tdf.timestamp,
        tdf.predicted,
        lw=2,
        color="tab:red",
        dashes=(10, 2),
        label="Trend line",
    )
    ax.grid(True, color="grey", lw=0.5, ls=":")
    tds_range = df.tds.max() - df.tds.min()
    ax.set_ylim(df.tds.min() - tds_range * 0.15, df.tds.max() + tds_range * 0.3)
    ax.legend(loc="lower center", frameon=False, fontsize="small", ncol=2)

    x0, x1 = pd.Timestamp(str(df.collected_date.dt.year.min() - 1)), pd.Timestamp(
        str(df.collected_date.dt.year.max() + 1)
    )
    ax.set_xlim(x0, x1)

    return ax


def plot_salinity_status_and_trend_simple_with_indicators(
    df, trend_df, trend_result, indicator_result, dfn, ax=None
):
    """Plot salinity status and trend results with indicator labels.

    Args:
        df (pd.DataFrame): containing data
        tdf (pd.DataFrame): containing the trend line data
        trend_result (pd.Series): containing the results of the calculation.
        indicator_result (pd.Series): row from dataframe containing the salinity
            indicator results e.g. columns "curr_tds_pct_diff", "curr_tds_pct_diff_indicator",
            "slope_pct_change_trend_pd", and "tds_trend_pct_change_indicator"
        dfn (pd.Series): row from the "Definitions_current_salinity_percent_diff"
            spreadsheet.
        ax (Axes): optional Axes object

    Returns: Axes object

    The data for this figure comes from the function :func:`get_historical_salinity_trends`,
    i.e. the last entry in the list of dataframes, and the last row of the
    "results" dataframe. It differs from the other method in that this only plots
    the data and trend, not the historical version of the trend.

    See also :func:`plot_salinity_status_and_trend_simple_with_indicators`

    """
    r = trend_result
    ax = plot_salinity_status_and_trend_simple(df, trend_df, r, ax=ax)
    ax.set_title(indicator_result.well_id)

    slabel = f"Current diff. to mean: {indicator_result.curr_tds_pct_diff:+.2f}%\n({indicator_result.curr_tds_pct_diff_indicator})"
    tlabel = f"Trend: {indicator_result.slope_pct_change_trend_pd:+.2f}% over {dfn.trend_length_years} yrs\n({indicator_result.tds_trend_pct_change_indicator})"
    ax.text(
        0.25,
        0.95,
        slabel,
        transform=ax.transAxes,
        color="white",
        bbox=dict(facecolor="tab:blue"),
        ha="center",
        va="top",
        fontsize="small",
    )
    ax.text(
        0.75,
        0.95,
        tlabel,
        transform=ax.transAxes,
        color="white",
        bbox=dict(facecolor="tab:red"),
        ha="center",
        va="top",
        fontsize="small",
    )

    ax.figure.tight_layout()
    return ax


def calculate_tds_bins(
    curr_sal=None,
    tds_bin_width=None,
    min_tds_colour=None,
    max_tds_colour=None,
    cmap=None,
    min_tds_bin=0,
    max_tds_bin=50000,
):
    """Generate colour bins for showing TDS data.

    Args:
        curr_sal (dict): dictionary with a key "tds" which stores an array
            of TDS values. If None, a synthetic one is generated which steps
            from the minimum to the maximum TDS specified, plus a high-salinity
            outlier.
        tds_bin_width (float): width of salinity bins.
        min_tds_colour ()

    """

    class Container:
        def summary(self):
            return {
                "cmap_name": self.cmap.name,
                "tds_bin_width": self.tds_bin_width,
                "min_tds_colour": self.min_tds_colour,
                "max_tds_colour": self.max_tds_colour,
                "min_tds_bin": self.min_tds_bin,
                "max_tds_bin": self.max_tds_bin,
            }

        def to_df(self):
            colours = [tuple(np.round(t, 3)) for t in self.bin_colours]
            r1 = np.asarray([t[0] for t in colours])
            g1 = np.asarray([t[1] for t in colours])
            b1 = np.asarray([t[2] for t in colours])
            alpha = np.array([t[3] for t in colours])
            r255 = r1 * 255
            g255 = g1 * 255
            b255 = b1 * 255
            rgb1 = [t[:3] for t in colours]
            rgba255 = [
                (int(c[0] * 255), int(c[1] * 255), int(c[2] * 255), c[3])
                for c in colours
            ]
            edges = [
                (self.bin_left_edges[i], self.bin_right_edges[i])
                for i in range(len(colours))
            ]
            return pd.DataFrame(
                {
                    "edges": edges,
                    "left_edge": self.bin_left_edges,
                    "right_edge": self.bin_right_edges,
                    "label": self.bin_labels,
                    "centre": self.bin_centre_tds,
                    "rgba1": colours,
                    "rgb1": rgb1,
                    "rgba255": rgba255,
                    "rgb255": [c[:3] for c in rgba255],
                    "alpha": alpha,
                    "r1": r1,
                    "g1": g1,
                    "b1": b1,
                    "r255": r255,
                    "g255": g255,
                    "b255": b255,
                }
            )

    if curr_sal is None:
        curr_sal = {"tds": [min_tds_bin, max_tds_bin, max_tds_bin + 100000]}

    c = Container()
    c.tds_bin_width = tds_bin_width
    c.min_tds_colour = min_tds_colour
    c.max_tds_colour = max_tds_colour
    c.min_tds_bin = min_tds_bin
    c.max_tds_bin = max_tds_bin
    c.cmap = cmap

    c.n_above_max = len([t for t in curr_sal["tds"] if t > c.max_tds_bin])

    logger.debug(
        f"tds_bin_width={tds_bin_width} / "
        f"min_tds_colour={min_tds_colour} / "
        f"max_tds_colour={max_tds_colour} / "
        f"max_tds_bin={max_tds_bin}"
    )

    c.all_bins = np.arange(c.min_tds_bin, c.max_tds_bin + 1, c.tds_bin_width)
    c.all_counts, c.all_bin_edges = np.histogram(curr_sal["tds"], bins=c.all_bins)
    c.all_bin_left_edges = c.all_bin_edges[:-1]
    logger.debug(f"all_bins: {[x for x in zip(range(len(c.all_bins)), c.all_bins)]}")
    logger.debug(
        f"all_counts: {[x for x in zip(range(len(c.all_counts)), c.all_counts)]}"
    )
    logger.debug(
        f"all_bin_edges: {[x for x in zip(range(len(c.all_bin_edges)), c.all_bin_edges)]}"
    )
    logger.debug(
        f"all_bin_left_edges: {[x for x in zip(range(len(c.all_bin_left_edges)), c.all_bin_left_edges)]}"
    )

    for i in range(len(c.all_counts)):
        count = c.all_counts[i]
        logger.debug(f"Looking from left (low) for non-zero counts at {i}")
        if count > 0:
            i0 = i
            logger.debug(f"Found {i} -> i0 = {i0}")
            break

    if c.n_above_max == 0:
        for i in range(len(c.all_counts))[::-1]:
            count = c.all_counts[i]
            logger.debug(f"Looking from right (high) for non-zero counts at {i}")
            if count > 0:
                i1 = i
                logger.debug(f"Found {i} -> i1 = {i1}")
                break
    else:
        i1 = len(c.all_counts) - 1
        logger.debug(
            f"n_above_max > 0 ({c.n_above_max}) therefore not trimming high bins (right-hand side) -> i1 = {i1}"
        )
    c.bins = c.all_bins[i0 : i1 + 1]
    c.bin_left_edges = c.all_bin_left_edges[i0 : i1 + 1]
    c.bin_right_edges = c.bin_left_edges + c.tds_bin_width
    c.counts = c.all_counts[i0 : i1 + 1]
    c.bin_centre_tds = [b + tds_bin_width / 2 for b in c.bin_left_edges]
    c.bin_labels = [
        f"{c.bin_left_edges[i]:.0f} to {(c.bin_left_edges[i] + tds_bin_width):.0f}"
        for i in range(len(c.bin_left_edges))
    ]
    logger.debug(f"Trimmed bins: {list(zip(range(len(c.bins)), c.bins))}")
    logger.debug(
        f"Trimmed bin_left_edges: {list(zip(range(len(c.bin_left_edges)), c.bin_left_edges))}"
    )
    logger.debug(f"Trimmed counts: {list(zip(range(len(c.counts)), c.counts))}")
    logger.debug(
        f"Trimmed bin_centre_tds: {list(zip(range(len(c.bin_centre_tds)), c.bin_centre_tds))}"
    )

    c.bin_tds_normed = [
        (c.bin_centre_tds[i] - min_tds_colour) / (max_tds_colour - min_tds_colour)
        for i in range(len(c.bin_centre_tds))
    ]
    c.bin_colours = [cmap(c.bin_tds_normed[i]) for i in range(len(c.bin_tds_normed))]

    if c.n_above_max > 0:
        c.bins = np.append(c.bins, c.max_tds_bin)
        c.bin_colours.append(tuple(list(mcolors.to_rgb("crimson")) + [1]))
        c.bin_labels.append(f"Above {c.max_tds_bin:.0f}")
        c.bin_left_edges = np.append(c.bin_left_edges, c.max_tds_bin)
        c.bin_right_edges = np.append(c.bin_right_edges, max(curr_sal["tds"]))
        c.all_counts = np.append(c.all_counts, c.n_above_max)
        c.all_bin_edges = np.append(c.all_bin_edges, max(curr_sal["tds"]))
        c.counts = np.append(c.counts, c.n_above_max)
        c.bin_centre_tds.append(np.nan)
        c.bin_tds_normed.append(np.nan)

    logger.debug(
        f"Final bin_left_edges: {list(zip(range(len(c.bin_left_edges)), c.bin_left_edges))}"
    )
    logger.debug(f"Final counts: {list(zip(range(len(c.counts)), c.counts))}")
    logger.debug(
        f"Final bin_centre_tds: {list(zip(range(len(c.bin_centre_tds)), c.bin_centre_tds))}"
    )

    return c


def plot_tds_current_bars(tdsbinned, ax=None, **kwargs):
    c = tdsbinned

    if ax is None:
        fig = plt.figure(figsize=(3.6, 2.8))
        ax = fig.add_subplot(111)

    x_counts = np.arange(len(c.counts))
    label_ypos = np.asarray(c.counts) + (max(c.counts) / 30)
    ax.barh(x_counts, c.counts, color=c.bin_colours, edgecolor="k")
    percentage_counts = [y / sum(c.counts) * 100 for y in c.counts]
    percentage_labels = [
        f"{pct:.0f}" for pct in round_to_100_percent(percentage_counts, 0)
    ]
    for i in range(len(x_counts)):
        if c.counts[i] > 0:
            ax.text(
                label_ypos[i],
                x_counts[i],
                percentage_labels[i] + "%",
                va="center",
                ha="left",
                color="#555555",
            )
    ax.set_yticks(np.arange(len(c.bin_labels)))
    ax.set_yticklabels(c.bin_labels)
    ax.set_ylim(len(c.bin_labels), -1)
    ax.set_ylabel("TDS (mg/L)")
    ax.set_xlabel("Number of wells")
    y0, y1 = ax.get_xlim()
    ax.xaxis.set_major_locator(
        mticker.MaxNLocator(nbins="auto", steps=[1, 2, 5, 10], integer=True)
    )
    ax.set_xlim(0, y1 + np.ceil((max(c.counts) / 30)))
    for sp in ["top", "right"]:
        ax.spines[sp].set_visible(False)
    return ax


def plot_tds_curr_pct_diff_bars(curr_pct_diffs, pct_diff, cm="auto", ax=None, **kwargs):
    if ax is None:
        fig = plt.figure(figsize=(3.6, 2.8))
        ax = fig.add_subplot(111)

    if cm == "auto":
        cm = PctDiffColormap(pct_diff)
    bin_labels = curr_pct_diffs.curr_tds_pct_diff_indicator.values
    counts = curr_pct_diffs.n_wells.fillna(0).values
    logger.debug(f"counts = {counts}")
    x_counts = np.arange(len(counts))
    label_ypos = np.asarray(counts) + (max(counts) / 30)
    ax.barh(x_counts, counts, color=cm.bin_colours, edgecolor="k")
    percentage_counts = curr_pct_diffs.pct_wells.fillna(0).values
    percentage_labels = [
        f"{pct:.0f}" for pct in round_to_100_percent(percentage_counts, 0)
    ]
    for i in range(len(x_counts)):
        if counts[i] > 0:
            ax.text(
                label_ypos[i],
                x_counts[i],
                percentage_labels[i] + "%",
                va="center",
                ha="left",
                color="#555555",
            )
    ax.set_yticks(np.arange(len(bin_labels)))
    ax.set_yticklabels(bin_labels)
    ax.set_ylim(len(bin_labels), -1)
    # ax.set_ylabel("TDS (mg/L)")
    ax.set_xlabel("Number of wells")
    y0, y1 = ax.get_xlim()
    ax.xaxis.set_major_locator(
        mticker.MaxNLocator(nbins="auto", steps=[1, 2, 5, 10], integer=True)
    )
    ax.set_xlim(0, y1 + np.ceil((max(counts) / 30)))
    for sp in ["top", "right"]:
        ax.spines[sp].set_visible(False)
    return ax


def plot_tds_trend_pct_change_bars(
    tds_trend_pct_changes, trend_pct, cm="auto", ax=None, **kwargs
):
    if ax is None:
        fig = plt.figure(figsize=(3.6, 2.8))
        ax = fig.add_subplot(111)

    if cm == "auto":
        cm = TrendPctColormap(trend_pct)
    bin_labels = tds_trend_pct_changes.tds_trend_pct_change_indicator.values
    counts = tds_trend_pct_changes.n_wells.fillna(0).values
    logger.debug(f"counts = {counts}")
    x_counts = np.arange(len(counts))
    label_ypos = np.asarray(counts) + (max(counts) / 30)
    ax.barh(x_counts, counts, color=cm.bin_colours, edgecolor="k")
    percentage_counts = tds_trend_pct_changes.pct_wells.fillna(0).values
    percentage_labels = [
        f"{pct:.0f}" for pct in round_to_100_percent(percentage_counts, 0)
    ]
    for i in range(len(x_counts)):
        if counts[i] > 0:
            ax.text(
                label_ypos[i],
                x_counts[i],
                percentage_labels[i] + "%",
                va="center",
                ha="left",
                color="#555555",
            )
    ax.set_yticks(np.arange(len(bin_labels)))
    ax.set_yticklabels(bin_labels)
    ax.set_ylim(len(bin_labels), -1)
    # ax.set_ylabel("TDS (mg/L)")
    ax.set_xlabel("Number of wells")
    y0, y1 = ax.get_xlim()
    ax.xaxis.set_major_locator(
        mticker.MaxNLocator(nbins="auto", steps=[1, 2, 5, 10], integer=True)
    )
    ax.set_xlim(0, y1 + np.ceil((max(counts) / 30)))
    for sp in ["top", "right"]:
        ax.spines[sp].set_visible(False)
    return ax


def plot_tds_current_map(
    tdsbinned,
    curr_sal,
    map_elements,
    map_annotations,
    ax=None,
    leg_frame=True,
    leg_loc="lower right",
    map_crs=8059,
    markersize=45,
    scattermarker="o",
    **kwargs,
):
    c = tdsbinned

    if ax is None:
        fig = plt.figure(figsize=(3.6, 2.8))
        ax = fig.add_subplot(111)

    for spine in ax.spines.values():
        spine.set_visible(False)

    legend = []
    if len(map_elements):
        if "legend_zorder" in map_elements[0]:
            map_elements = sorted(map_elements, key=lambda x: x["legend_zorder"])
    for me in map_elements:
        me["shapely_object"].plot(ax=ax, **me["plot_kwargs"])
        if me.get("label", None):
            if me["legend_item_type"] == "line":
                legend_patch = ax.plot([], [], **me["legend_kwargs"])[0]
            elif me["legend_item_type"] == "rect":
                legend_patch = mpatches.Rectangle((0, 0), 1, 1, **me["legend_kwargs"])
            legend.append((me["label"], legend_patch))

    # bins_for_cut =  np.append(c.bins, c.bins[-1] + c.tds_bin_width)
    # logger.debug(f"bins_for_cut {bins_for_cut}")
    well_bin_lefts = np.array(
        [
            np.nan if i is np.nan else int(i.left)
            for i in pd.cut(curr_sal.tds, c.bin_left_edges, right=False)
        ]
    )
    well_bin_lefts = []
    for ix, row in curr_sal.iterrows():
        interval = False
        logger.debug(f"testing {row.well_id} {row.tds} against bins...")
        for i in range(len(c.bins)):
            b_left = c.bin_left_edges[i]
            b_right = c.bin_right_edges[i]
            if row.tds > b_left and row.tds <= b_right:
                interval = b_left
                break
        if interval is False:
            interval = np.nan
        well_bin_lefts.append(interval)

    logger.debug(
        "\n".join([str(x) for x in zip(curr_sal.well_id, curr_sal.tds, well_bin_lefts)])
    )

    tr = pyproj.Transformer.from_crs(7844, map_crs, always_xy=True)

    for i in range(len(c.bin_left_edges)):
        bin_left = c.bin_left_edges[i]
        idx = well_bin_lefts == bin_left
        lon = curr_sal[idx].longitude.values
        lat = curr_sal[idx].latitude.values
        logger.debug(
            f"Mapping {curr_sal[idx].well_id.values} for bin_left_edge {bin_left} ( lons = {lon}    lats = {lat}"
        )
        if len(lon):
            easting, northing = tr.transform(lon, lat)
            ax.scatter(
                easting,
                northing,
                marker=scattermarker,
                color=c.bin_colours[i],
                edgecolor="#555555",
                linewidth=0.3,
                s=markersize,
                zorder=5,
            )

    map_final_annotations = []
    for ann in map_annotations:
        if not ann["well_id"] in curr_sal.well_id.unique():
            logger.warning(
                f"****Warning: annotation {ann} is not in the dataset - cannot add it ******"
            )
        else:
            map_final_annotations.append(ann)

    for ann in map_final_annotations:
        row = curr_sal[curr_sal.well_id == ann["well_id"]].iloc[0]
        label = ann.get("label", ann["well_id"])
        easting, northing = tr.transform(row.longitude, row.latitude)
        arrowprops = ann.get("arrowprops", None)
        if arrowprops is not None:
            arrowprops = dict(
                {
                    "arrowstyle": "-|>",
                    "color": "#444444",
                    "mutation_scale": 8,
                    "shrinkA": 0,
                    "shrinkB": 0,
                },
                **arrowprops,
            )
        ax.annotate(
            label,
            (easting, northing),
            (easting + ann["x_km"] * 1000, northing + ann["y_km"] * 1000),
            color="#444444",
            fontsize="x-small",  # map annotation
            ha=ann["ha"],
            va=ann["va"],
            arrowprops=arrowprops,
            zorder=6,
        )
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_xticklabels([])
    ax.set_yticklabels([])

    if len(legend):
        labels, patches = zip(*legend)
        if leg_frame:
            frameon = True
            framealpha = 1
            edgecolor = "white"
        else:
            frameon = False
            framealpha = None
            edgecolor = None
        leg = ax.legend(
            patches,
            labels,
            loc=leg_loc,
            fontsize="x-small",
            frameon=frameon,
            framealpha=framealpha,
            edgecolor=edgecolor,
        )
        for t in leg.get_texts():
            t.set_color("#444444")
    return ax


def plot_tds_curr_pct_diff_map(
    curr_pct_diff,
    pct_diff,
    map_elements,
    map_annotations,
    cm="auto",
    label_col="curr_tds_pct_diff_indicator",
    ax=None,
    leg_frame=True,
    leg_loc="lower right",
    map_crs=8059,
    markersize=45,
    scattermarker="s",
    **kwargs,
):
    if ax is None:
        fig = plt.figure(figsize=(3.6, 2.8))
        ax = fig.add_subplot(111)

    for spine in ax.spines.values():
        spine.set_visible(False)

    legend = []
    if len(map_elements):
        if "legend_zorder" in map_elements[0]:
            map_elements = sorted(map_elements, key=lambda x: x["legend_zorder"])
    for me in map_elements:
        me["shapely_object"].plot(ax=ax, **me["plot_kwargs"])
        if me.get("label", None):
            if me["legend_item_type"] == "line":
                legend_patch = ax.plot([], [], **me["legend_kwargs"])[0]
            elif me["legend_item_type"] == "rect":
                legend_patch = mpatches.Rectangle((0, 0), 1, 1, **me["legend_kwargs"])
            legend.append((me["label"], legend_patch))

    if cm == "auto":
        cm = PctDiffColormap(pct_diff)

    tr = pyproj.Transformer.from_crs(7844, map_crs, always_xy=True)
    logger.debug(f"curr_pct_diff = \n{curr_pct_diff[['well_id', label_col]]}")

    for i in range(len(pct_diff["bins_left"])):
        label = pct_diff["labels"][i]
        idx = curr_pct_diff[label_col] == label
        lon = curr_pct_diff[idx].longitude.values
        lat = curr_pct_diff[idx].latitude.values
        if len(lon):
            easting, northing = tr.transform(lon, lat)
            ax.scatter(
                easting,
                northing,
                marker=scattermarker,
                color=cm.label_colours[i],
                edgecolor="#555555",
                linewidth=0.3,
                s=markersize,
                zorder=5,
            )

    map_final_annotations = []
    for ann in map_annotations:
        if not ann["well_id"] in curr_pct_diff.well_id.unique():
            logger.warning(
                f"****Warning: annotation {ann} is not in the dataset - cannot add it ******"
            )
        else:
            map_final_annotations.append(ann)

    for ann in map_final_annotations:
        row = curr_pct_diff[curr_pct_diff.well_id == ann["well_id"]].iloc[0]
        label = ann.get("label", ann["well_id"])
        easting, northing = tr.transform(row.longitude, row.latitude)
        arrowprops = ann.get("arrowprops", None)
        if arrowprops is not None:
            arrowprops = dict(
                {
                    "arrowstyle": "-|>",
                    "color": "#444444",
                    "mutation_scale": 8,
                    "shrinkA": 0,
                    "shrinkB": 0,
                },
                **arrowprops,
            )
        ax.annotate(
            label,
            (easting, northing),
            (easting + ann["x_km"] * 1000, northing + ann["y_km"] * 1000),
            color="#444444",
            fontsize="x-small",  # map annotation
            ha=ann["ha"],
            va=ann["va"],
            arrowprops=arrowprops,
            zorder=6,
        )
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_xticklabels([])
    ax.set_yticklabels([])

    if len(legend):
        labels, patches = zip(*legend)
        if leg_frame:
            frameon = True
            framealpha = 1
            edgecolor = "white"
        else:
            frameon = False
            framealpha = None
            edgecolor = None
        leg = ax.legend(
            patches,
            labels,
            loc=leg_loc,
            fontsize="x-small",
            frameon=frameon,
            framealpha=framealpha,
            edgecolor=edgecolor,
        )
        for t in leg.get_texts():
            t.set_color("#444444")
    return ax


def plot_tds_trend_pct_change_map(
    tds_trend_pct_changes,
    trend_pct,
    map_elements,
    map_annotations,
    cm="auto",
    label_col="tds_trend_pct_change_indicator",
    ax=None,
    leg_frame=True,
    leg_loc="lower right",
    map_crs=8059,
    markersize=45,
    increase_scattermarker="^",
    decrease_scattermarker="v",
    **kwargs,
):
    if ax is None:
        fig = plt.figure(figsize=(3.6, 2.8))
        ax = fig.add_subplot(111)

    for spine in ax.spines.values():
        spine.set_visible(False)

    legend = []
    if len(map_elements):
        if "legend_zorder" in map_elements[0]:
            map_elements = sorted(map_elements, key=lambda x: x["legend_zorder"])
    for me in map_elements:
        me["shapely_object"].plot(ax=ax, **me["plot_kwargs"])
        if me.get("label", None):
            if me["legend_item_type"] == "line":
                legend_patch = ax.plot([], [], **me["legend_kwargs"])[0]
            elif me["legend_item_type"] == "rect":
                legend_patch = mpatches.Rectangle((0, 0), 1, 1, **me["legend_kwargs"])
            legend.append((me["label"], legend_patch))

    if cm == "auto":
        cm = TrendPctColormap(trend_pct)

    tr = pyproj.Transformer.from_crs(7844, map_crs, always_xy=True)
    logger.debug(
        f"tds_trend_pct_changes = \n{tds_trend_pct_changes[['well_id', label_col]]}"
    )

    for i in range(len(trend_pct["bins_left"])):
        label = trend_pct["labels"][i]
        if "increase" in label:
            scattermarker = increase_scattermarker
        elif "decrease" in label:
            scattermarker = decrease_scattermarker
        idx = tds_trend_pct_changes[label_col] == label
        lon = tds_trend_pct_changes[idx].longitude.values
        lat = tds_trend_pct_changes[idx].latitude.values
        if len(lon):
            easting, northing = tr.transform(lon, lat)
            ax.scatter(
                easting,
                northing,
                marker=scattermarker,
                color=cm.label_colours[i],
                edgecolor="#555555",
                linewidth=0.3,
                s=markersize,
                zorder=5,
            )

    map_final_annotations = []
    for ann in map_annotations:
        if not ann["well_id"] in tds_trend_pct_changes.well_id.unique():
            logger.warning(
                f"****Warning: annotation {ann} is not in the dataset - cannot add it ******"
            )
        else:
            map_final_annotations.append(ann)

    for ann in map_final_annotations:
        row = tds_trend_pct_changes[
            tds_trend_pct_changes.well_id == ann["well_id"]
        ].iloc[0]
        label = ann.get("label", ann["well_id"])
        easting, northing = tr.transform(row.longitude, row.latitude)
        arrowprops = ann.get("arrowprops", None)
        if arrowprops is not None:
            arrowprops = dict(
                {
                    "arrowstyle": "-|>",
                    "color": "#444444",
                    "mutation_scale": 8,
                    "shrinkA": 0,
                    "shrinkB": 0,
                },
                **arrowprops,
            )
        ax.annotate(
            label,
            (easting, northing),
            (easting + ann["x_km"] * 1000, northing + ann["y_km"] * 1000),
            color="#444444",
            fontsize="x-small",  # map annotation
            ha=ann["ha"],
            va=ann["va"],
            arrowprops=arrowprops,
            zorder=6,
        )
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_xticklabels([])
    ax.set_yticklabels([])

    if len(legend):
        labels, patches = zip(*legend)
        if leg_frame:
            frameon = True
            framealpha = 1
            edgecolor = "white"
        else:
            frameon = False
            framealpha = None
            edgecolor = None
        leg = ax.legend(
            patches,
            labels,
            loc=leg_loc,
            fontsize="x-small",
            frameon=frameon,
            framealpha=framealpha,
            edgecolor=edgecolor,
        )
        for t in leg.get_texts():
            t.set_color("#444444")
    return ax


def plot_tds_trend_triclass_bars(sal_trends, ax=None):
    gsr_colours = status_change_colours
    gsr_colours_rgba = {
        k: np.asarray([x / 255 for x in rgb] + [1]) for k, rgb in gsr_colours.items()
    }

    status_key_x = [0, 1, 2]
    status_keys = ["Increasing", "Stable", "Decreasing"]
    status_values = [
        len(sal_trends[sal_trends.status_change == v]) for v in status_keys
    ]
    status_colours = [gsr_colours_rgba[s] for s in status_keys]

    if ax is None:
        fig = plt.figure(figsize=(3.6, 2.0))
        ax = fig.add_subplot(111)

    ax.barh(
        status_key_x,
        status_values,
        height=0.6,
        color=status_colours,
        edgecolor="#555555",
    )
    ax.set_yticks(status_key_x)
    ax.set_yticklabels(status_keys)  # rotation=40, ha="right")
    ax.set_ylim(status_key_x[-1] + 1, status_key_x[0] - 1)
    # ax.set_ylabel("Five year change in TDS (%)", fontsize="medium")

    label_ypos = np.asarray(status_values) + (max(status_values) / 30)
    percentage_counts = [y / sum(status_values) * 100 for y in status_values]
    percentage_labels = [
        f"{pct:.0f}" for pct in round_to_100_percent(percentage_counts, 0)
    ]
    for i, x in enumerate(status_key_x):
        if status_values[i] > 0:
            ax.text(
                label_ypos[i],
                x,
                percentage_labels[i] + "%",
                va="center",
                ha="left",
                color="#555555",
            )
    ax.set_xlabel("Number of wells", fontsize="medium")

    max_counts = max(status_values)
    ax.set_xlim(0, max_counts + (max_counts / 2.0))
    ax.xaxis.set_major_locator(mticker.MaxNLocator(integer=True))

    for sp in ("top", "right"):
        ax.spines[sp].set_visible(False)
    return ax


def plot_tds_trend_triclass_map(
    sal_trends,
    map_elements,
    map_annotations,
    ax=None,
    leg_frame=True,
    map_crs=8059,
    leg_loc="lower right",
    markersize=60,
):
    tds_trend_map_markers = {"Increasing": "^", "Stable": "s", "Decreasing": "v"}

    gsr_colours = status_change_colours
    gsr_colours_rgba = {
        k: np.asarray([x / 255 for x in rgb] + [1]) for k, rgb in gsr_colours.items()
    }

    if ax is None:
        fig = plt.figure(figsize=(3.6, 2.8))
        ax = fig.add_subplot(111)

    for spine in ax.spines.values():
        spine.set_visible(False)

    legend = []
    if len(map_elements):
        if "legend_zorder" in map_elements[0]:
            map_elements = sorted(map_elements, key=lambda x: x["legend_zorder"])
    for me in map_elements:
        me["shapely_object"].plot(ax=ax, **me["plot_kwargs"])
        if me.get("label", None):
            if me["legend_item_type"] == "line":
                legend_patch = ax.plot([], [], **me["legend_kwargs"])[0]
            elif me["legend_item_type"] == "rect":
                legend_patch = mpatches.Rectangle((0, 0), 1, 1, **me["legend_kwargs"])
            legend.append((me["label"], legend_patch))

    tr = pyproj.Transformer.from_crs(7844, map_crs, always_xy=True)
    for trend_label in ["Increasing", "Stable", "Decreasing"]:
        dfg = sal_trends[sal_trends.status_change == trend_label]
        if len(dfg):
            easting, northing = tr.transform(dfg.longitude.values, dfg.latitude.values)
            ax.scatter(
                easting,
                northing,
                marker=tds_trend_map_markers[trend_label],
                color=gsr_colours_rgba[trend_label],
                edgecolor="#555555",
                linewidth=0.3,
                s=markersize,
                zorder=5,
            )

    map_final_annotations = []
    for ann in map_annotations:
        if not ann["well_id"] in sal_trends.well_id.unique():
            logger.warning(
                f"****Warning: annotation {ann} is not in the dataset - cannot add it ******"
            )
        else:
            map_final_annotations.append(ann)

    for ann in map_final_annotations:
        row = sal_trends[sal_trends.well_id == ann["well_id"]].iloc[0]
        label = ann.get("label", ann["well_id"])
        easting, northing = tr.transform(row.longitude, row.latitude)
        arrowprops = ann.get("arrowprops", None)
        if arrowprops is not None:
            arrowprops = dict(
                {"arrowstyle": "->", "color": "#444444", "mutation_scale": 10},
                **arrowprops,
            )
        ax.annotate(
            label,
            (easting, northing),
            (easting + ann["x_km"] * 1000, northing + ann["y_km"] * 1000),
            color="#444444",
            fontsize="x-small",  # map annotation
            ha=ann["ha"],
            va=ann["va"],
            arrowprops=arrowprops,
            zorder=6,
        )
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_xticklabels([])
    ax.set_yticklabels([])

    if len(legend):
        labels, patches = zip(*legend)
        if leg_frame:
            frameon = True
            framealpha = 1
            edgecolor = "white"
        else:
            frameon = False
            framealpha = None
            edgecolor = None
        leg = ax.legend(
            patches,
            labels,
            loc=leg_loc,
            fontsize="x-small",
            frameon=frameon,
            framealpha=framealpha,
            edgecolor=edgecolor,
        )
        for t in leg.get_texts():
            t.set_color("#444444")
    return ax


def plot_salinity_curr_pct_diffs_summary(
    curr_pct_diffs, pct_diff, results=None, cm="auto", style="internal", ax=None
):
    """Show summary figure for current TDS percentage differences.

    Args:
        curr_pct_diffs (pd.Series): from code
        pct_diff (dict): bins
        results (pd.DataFrame, optional): list of wells with curr_tds_pct_diff column. Optional.
        style (str): either 'internal' or 'publication'
        ax (Axes): optional
        cm (custom SalinityColormap object): or "auto" which will use
            :class:`wrap_technote.PctDiffColormap`.

    Returns: Axes object.

    """
    if style == "internal":
        ytick_fontsize = "small"
        default_figsize = (7, 4)
    elif style == "publication":
        ytick_fontsize = "medium"
        default_figsize = (5, 4)

    if cm == "auto":
        cm = PctDiffColormap(pct_diff)
    if ax is None:
        fig = plt.figure(figsize=default_figsize)
        ax = fig.add_subplot(111)
    y_idx = np.arange(len(curr_pct_diffs))
    ax.barh(
        y_idx,
        curr_pct_diffs.n_wells,
        color=cm.bin_colours,
        edgecolor="gray",
        linewidth=0.6,
    )
    ax.set_ylim(y_idx[-1] + 1, y_idx[0] - 1)
    _ = ax.set_yticks(y_idx)
    _ = ax.set_yticklabels(
        curr_pct_diffs.curr_tds_pct_diff_indicator, fontsize=ytick_fontsize
    )
    ax.set_xlabel("Number of wells")
    if style == "publication":
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
    elif style == "internal":
        median_class = get_median_class(
            curr_pct_diffs.n_wells,
            class_values=curr_pct_diffs.curr_tds_pct_diff_indicator.values,
        )
        for i in y_idx:
            n = curr_pct_diffs.n_wells.iloc[i]
            pct = curr_pct_diffs.pct_wells.iloc[i]
            text = f"{n:.0f} ({pct:.1f}%)"
            ax.text(n, i, text, fontsize="small")
        label = f'Medians: class="{median_class}"'
        if not results is None:
            median_value = results.curr_tds_pct_diff.dropna().median()
            label += f" value={median_value:+.2f}%"
        ax.set_title(label, fontsize="small")
        x0, x1 = ax.get_xlim()
        ax.set_xlim(x0, x1 + (x1 - x0) / 10)
    ax.figure.tight_layout()
    return ax


def plot_salinity_trend_pct_summary(
    tds_trend_pct_changes, trend_pct, results=None, cm="auto", style="internal", ax=None
):
    """Show summary figure for current TDS percentage differences.

    Args:
        tds_trend_pct_changes (pd.Series): from code
        trend_pct (dict): bins
        results (pd.DataFrame, optional): list of wells with slope_pct_change_trend_pd column. Optional.
        cm (custom SalinityColormap object): or "auto" which will use
            :class:`wrap_technote.TrendPctColormap`.
        style (str): either 'internal' or 'publication'
        ax (Axes): optional

    Returns: Axes object.

    """
    if style == "internal":
        ytick_fontsize = "small"
        default_figsize = (7, 4)
    elif style == "publication":
        ytick_fontsize = "medium"
        default_figsize = (5, 4)
    if cm == "auto":
        cm = TrendPctColormap(trend_pct)
    if ax is None:
        fig = plt.figure(figsize=default_figsize)
        ax = fig.add_subplot(111)
    y_idx = np.arange(len(tds_trend_pct_changes))
    ax.barh(
        y_idx,
        tds_trend_pct_changes.n_wells,
        color=cm.bin_colours,
        edgecolor="gray",
        linewidth=0.6,
    )
    ax.set_ylim(y_idx[-1] + 1, y_idx[0] - 1)
    _ = ax.set_yticks(y_idx)
    _ = ax.set_yticklabels(
        tds_trend_pct_changes.tds_trend_pct_change_indicator, fontsize=ytick_fontsize
    )
    ax.set_xlabel("Number of wells")
    if style == "publication":
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
    elif style == "internal":
        median_class = get_median_class(
            tds_trend_pct_changes.n_wells,
            class_values=tds_trend_pct_changes.tds_trend_pct_change_indicator,
        )
        for i in y_idx:
            n = tds_trend_pct_changes.n_wells.iloc[i]
            pct = tds_trend_pct_changes.pct_wells.iloc[i]
            text = f"{n:.0f} ({pct:.1f}%)"
            ax.text(n, i, text, fontsize="small")
        label = f'Medians: class="{median_class}"'
        if not results is None:
            median_value = results.slope_pct_change_trend_pd.dropna().median()
            label += f" value={median_value:+.2f}%"
        ax.set_title(label, fontsize="small")
    ax.figure.tight_layout()
    return ax


class SalinityBinsColormap(object):
    def __init__(self, bins_dfn, cmap, cmap_r):
        self.bins_dfn = bins_dfn
        self.labels = self.bins_dfn["labels"]
        self.split_labels = [
            l.replace("% above", "%\nabove").replace("% below", "%\nbelow")
            for l in self.labels
        ]

        self.norm_values = mcolors.Normalize(
            self.bins_dfn["bins"][1], self.bins_dfn["bins"][-2]
        )
        self.norm_label_indices = mcolors.Normalize(0, len(self.labels) - 1)

        self.cmap = copy.copy(plt.cm.get_cmap(cmap))
        self.cmap_values = copy.copy(plt.cm.get_cmap(cmap))
        self.cmap_label_indices = copy.copy(plt.cm.get_cmap(cmap_r))

        self.cmap.set_bad("darkgrey")
        self.cmap_values.set_bad("darkgrey")
        self.cmap_label_indices.set_bad("darkgrey")

        self.bin_colours = [
            self.cmap(self.norm_values(v)) for v in self.bins_dfn["bin_centres"]
        ]
        self.label_colours = [
            self.cmap_label_indices(self.norm_label_indices(i))
            for i in range(len(self.labels))
        ]
        self.label_to_index = dict(zip(self.labels, range(len(self.labels))))
        self.label_to_colour = dict(zip(self.labels, self.bin_colours))


class PctDiffColormap(SalinityBinsColormap):
    def __init__(self, bins_dfn):
        super().__init__(bins_dfn, "PiYG", "PiYG_r")


class TrendPctColormap(SalinityBinsColormap):
    def __init__(self, bins_dfn):
        super().__init__(bins_dfn, "BrBG", "BrBG_r")


def plot_salinity_historical_pct_diffs(
    tdsann,
    pct_diff,
    tds_indicator_col="tds_indicator",
    cm="auto",
    included_wells=None,
):
    """Plot historical salinity percent differences.

    Args:
        tdsann (pd.DataFrame): annual salinities - must have columns "well_id",
            "tds", "collected_year", and *tds_indicator_col* e.g. "tds_indicator"
        pct_diff (dict): definition of salinity percent difference bins
        tds_indicator_col (str): name of column in *tdsann* with the pct_diff indicator
            label
        cm (SalinityColormap, str): use "auto" to use PctDiffColormap

    """
    if cm == "auto":
        cm = PctDiffColormap(pct_diff)

    year0 = tdsann.collected_year.min()
    year1 = tdsann.collected_year.max()
    years = np.arange(year0, year1 + 1, 1)
    all_well_ids = sorted(tdsann.well_id.unique())
    if included_wells is not None:
        well_ids = list(included_wells) + [
            w for w in all_well_ids if not w in included_wells
        ]
    else:
        well_ids = all_well_ids
    nx = len(years)
    ny = len(well_ids)
    arr = np.empty((nx, ny), dtype=int) * np.nan
    for j in range(ny):
        well_id = well_ids[j]
        s = (
            tdsann[(tdsann.well_id == well_id)]
            .set_index("collected_year")[tds_indicator_col]
            .reindex(years)
        )
        arr[:, j] = s.map(cm.label_to_index).values

    fig_x = nx / 8
    fig_y = ny / 8
    fig_y2 = fig_y + 1.5
    min_fig_y = 5
    if fig_y2 < min_fig_y:
        fig_y2 = min_fig_y
    fig = plt.figure(figsize=(fig_x, fig_y2))
    gs = gridspec.GridSpec(2, 1, height_ratios=(20, 1))
    ax = fig.add_subplot(gs[0])
    cax = fig.add_subplot(gs[1])
    axim = ax.imshow(arr.T, cmap=cm.cmap_label_indices, aspect="auto")
    for j in range(ny):
        ax.axhline(j - 0.5, color="grey", lw=0.3)

    if included_wells is not None:
        ax.axhline(len(included_wells) - 0.5, color="green", lw=1.5)
    cbar = plot_colorbar(
        cax=cax,
        colours=cm.label_colours,
        cbar_labels=cm.split_labels,
        orientation="horizontal",
    )
    cbar.ax.tick_params(labelsize="small")
    plt.setp(cbar.ax.get_xticklabels(), rotation=30, ha="right")
    for i, y in enumerate(years):
        if y % 10 == 0:
            ax.axvline(i - 0.5, lw=0.5, color="k")
        elif y % 5 == 0:
            ax.axvline(i - 0.5, lw=0.5, dashes=(10, 2), color="k")
        else:
            ax.axvline(i - 0.5, lw=0.3, color="grey")

    # ax.set_ylim(*ax.get_ylim()[::-1])
    _ = ax.set_xticks(range(len(years)))
    _ = ax.set_xticklabels(
        [str(y) for y in years], rotation=90, fontsize="x-small", ha="center"
    )
    _ = ax.set_yticks(range(len(well_ids)))
    _ = ax.set_yticklabels(well_ids, fontsize="x-small")
    ax.tick_params(
        "y", which="major", left=True, right=True, labelleft=True, labelright=True
    )
    ax.tick_params(
        "x", which="major", top=True, bottom=True, labeltop=True, labelbottom=True
    )
    fig.tight_layout()
    return ax, cax



import colorsys
from datetime import date, datetime, timedelta
import io
import re
import os
import logging
import textwrap
from pathlib import Path

import matplotlib.pyplot as plt
from matplotlib import lines as mlines
from matplotlib import gridspec
from matplotlib import dates as mdates
from matplotlib import ticker as mticker
from matplotlib import colors as mcolors
import numpy as np
import pandas as pd
from adjustText import adjust_text
from scipy import stats
from PIL import Image, ImageChops
import pyproj

from .gwutils import *
from .utils import *


logger = get_logger()

from .charts_utils import *
from .charts_wl import *
from .charts_tds import *
from .charts_rf import *



from pkg_resources import get_distribution, DistributionNotFound

try:
    __version__ = get_distribution(__name__).version
except DistributionNotFound:
    # package is not installed
    pass


from wrap_technote.utils import *
from wrap_technote.gwutils import *
from wrap_technote.charts import *
from wrap_technote.fileio import *
from wrap_technote.reporting import *
from wrap_technote.salinity import *
from wrap_technote.waterlevels import *
from wrap_technote.rainfall import *

from wrap_technote import scripts


from pathlib import Path

import warnings

warnings.filterwarnings(
    "ignore",
    module="wrap_technote.reporting",
    # lineno=800
)



import click

from .waterlevels import waterlevels, waterlevels_run
from .salinities import salinities, salinities_run
from .summaries import summaries, summaries_run
from .rainfall import rainfall

from .. import __version__, Report, CURRENT_RPERIOD
from ..utils import get_logger

logger = get_logger()


@click.group()
def wraptn():
    # bump commit
    pass


@click.command()
@click.option("-p", "--reporting-period", default=CURRENT_RPERIOD)
@click.option("-v", "--verbose", count=True)
@click.option("-s1", "--step1/--no-step1", default=True)
@click.option("-s2", "--step2/--no-step2", default=True)
@click.option("-s3", "--step3/--no-step3", default=True)
@click.option("-s4", "--step4/--no-step4", default=True)
@click.option("-s5", "--step5/--no-step5", default=True)
@click.option("-s6", "--step6/--no-step6", default=True)
@click.option("-s7", "--step7/--no-step6", default=True)
@click.option("-s", "--static/--no-static", default=True)
@click.option("--nbs/--no-nbs", default=True)
@click.option("--local/--no-local", default=True)
@click.option("-r", "--report", default="")
@click.option("--figures/--no-figures", default=True)
@click.argument("resource", required=False)
def run(
    resource,
    reporting_period,
    verbose,
    step1,
    step2,
    step3,
    step4,
    step5,
    step6,
    step7,
    static,
    nbs,
    local,
    report,
    figures,
):
    handlers = []
    if verbose == 1:
        handlers.append({"sink": sys.stdout, "level": "INFO"})
    if verbose == 2:
        handlers.append({"sink": sys.stdout, "level": "INFO"})
        handlers.append({"sink": "wraptn_cli.log", "level": "DEBUG"})
    config = {
        "handlers": handlers,
    }
    logger.configure(**config)

    logger.warning(f"wraptn.__version__ = {__version__}")
    wl_resource_keys = []
    tds_resource_keys = []
    if resource is None:
        r = Report(report, reporting_period)
        df = r.read_table("Report_Resources_mapping")
        wl_resource_keys = [r for r in df[df.param == "WL"].resource_key.unique() if r]
        tds_resource_keys = [
            r for r in df[df.param == "TDS"].resource_key.unique() if r
        ]
    else:
        if resource.endswith("_WL"):
            wl_resource_keys.append(resource)
        if resource.endswith("_TDS"):
            tds_resource_keys.append(resource)
    logger.info(
        f"CLI will run WL resources: {wl_resource_keys} and TDS resources: {tds_resource_keys}"
    )
    for rk in wl_resource_keys:
        waterlevels_run(
            rk,
            reporting_period,
            verbose,
            step1,
            step2,
            step3,
            static,
            report,
            figures,
        )
    for rk in tds_resource_keys:
        salinities_run(
            rk,
            reporting_period,
            verbose,
            step4,
            step5,
            step6,
            step7,
            static,
            report,
            figures,
        )
    if report:
        summaries_run(reporting_period, report, verbose, resource, local, nbs)


wraptn.add_command(waterlevels)
wraptn.add_command(salinities)
wraptn.add_command(summaries)
wraptn.add_command(rainfall)
wraptn.add_command(run)
